import numpy as np
import os
import re
import json
from .xrd import cal_UnitCellVolume, cal_dspacing
from .jcpds_dioptas import jcpds, jcpds_reflection
import datetime
from scipy import interpolate

# import numpy.ma as ma

try:
    import pymatgen.core as mg
    from pymatgen.analysis.diffraction.xrd import XRDCalculator
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
except ImportError:
    mg = None
    XRDCalculator = None
    SpacegroupAnalyzer = None


def bm3_p(volume, v0, k0, k0p):
    volume = np.asarray(volume, dtype=float)
    eta = np.power(v0 / volume, 1.0 / 3.0)
    pressure = 1.5 * k0 * (np.power(eta, 7.0) - np.power(eta, 5.0))
    pressure *= 1.0 + 0.75 * (k0p - 4.0) * (np.power(eta, 2.0) - 1.0)
    if pressure.ndim == 0:
        return float(pressure)
    return pressure


def bm3_v(pressure, v0, k0, k0p, min_strain=0.3):
    pressure = float(pressure)
    if pressure <= 0.0:
        return float(v0)

    min_volume = float(v0) * max(1.0e-6, (1.0 - float(min_strain)) ** 3)
    low = min_volume
    high = float(v0)

    p_low = float(bm3_p(low, v0, k0, k0p))
    while p_low < pressure and low > v0 * 1.0e-6:
        low *= 0.5
        p_low = float(bm3_p(low, v0, k0, k0p))

    for __ in range(80):
        mid = 0.5 * (low + high)
        p_mid = float(bm3_p(mid, v0, k0, k0p))
        if abs(p_mid - pressure) < 1.0e-8:
            return mid
        if p_mid > pressure:
            low = mid
        else:
            high = mid
    return 0.5 * (low + high)


def _require_pymatgen_for_cif():
    if mg is not None and XRDCalculator is not None and SpacegroupAnalyzer is not None:
        return
    raise ImportError(
        "CIF import requires the optional 'pymatgen' dependency.")


class DiffractionLine:
    """
    Class that defines a single diffraction line.
    Attributes:
      dsp0: d-spacing at reference conditions (P = 0 GPa and T = 300 K)
      dsp:  d-spacing at high PT
      intensity:  diffraction intensity read from the file
      h, k, l: Miller index
    """

    def __init__(self):
        self.dsp0 = 0.
        self.dsp = 0.
        self.intensity = 0.
        self.h = 0.
        self.k = 0.
        self.l = 0.
        self.label = ''
        self.pressure_coeffs = []
        self.temperature_coeffs = []


class UnitCell(object):
    """
    Class that defines a unit cell properties
    """

    def __init__(self):
        self.name = ''
        self.a = 0.
        self.b = 0.
        self.c = 0.
        self.alpha = 0.
        self.beta = 0.
        self.gamma = 0.
        self.v = 0.
        self.DiffLines = []
        self.symmetry = ''
        self.color = ''
        self.display = True

    def cal_dsp(self):
        """
        b_a and c_a are newly included for adjusting axial ratios.
        For cubic structure, these two inputs are ignored.
        For tetragonal and hexagonal, only c_a will be used.
        """
#            a = self.a; b = self.b; c = self.c;
#            alpha = self.alpha; beta = self.beta; gamma = self.gamma
        self.v = cal_UnitCellVolume(self.symmetry, self.a, self.b, self.c,
                                    self.alpha, self.beta, self.gamma)
        for dl in self.DiffLines:
            dsp = cal_dspacing(self.symmetry, dl.h, dl.k, dl.l,
                               self.a, self.b, self.c,
                               self.alpha, self.beta, self.gamma)
            dl.dsp = dsp

    def get_tthVSint(self, wavelength):
        """
        Returns twoth and intensity for bar plots in PyPeakPo
        Note that tth and intensity are numpy array not list
        If P, T, b_a, c_a have changed, run cal_dsp first for update
        """
#        self.cal_dsp(pressure, temperature, b_a, c_a)
        dsp = []
        intensity = []
        for line in self.DiffLines:
            d = line.dsp
            i = line.intensity
            dsp.append(d)
            intensity.append(i)
        tth = 2. * np.degrees(np.arcsin(wavelength / 2. / np.array(dsp)))
        return tth, np.array(intensity)

    def find_DiffLine(self, tth_c, wavelength):
        """
        Returns index of the cloest diffraction line, its difference
            in abs tth
        If P, T, b_a, c_a have changed, run cal_dsp first for update
        """
        tth, intensity = self.get_tthVSint(wavelength)

        idx = (np.abs(tth - tth_c)).argmin()

        return idx, abs(tth[idx] - tth_c), tth[idx]

    def make_TextOutput(self, pressure, temperature):
        if self.phase_kind == 'rapo':
            textout = self.name + '\n'
            textout += 'RAPO Raman mode list\n'
            textout += 'Below are the modes at {0: 6.2f} GPa, {1: 5.1f} K\n'.format(
                pressure, temperature)
            textout += 'mode, shift (cm-1), intensity\n'
            for dl in self.DiffLines:
                label = getattr(dl, 'label', 'mode')
                textout += '{0}, {1: .5f}, {2: .1f}\n'.format(
                    label, float(dl.dsp), float(dl.intensity))
            return textout
        textout = self.name + '\n'
        textout += self.symmetry + '\n'
        textout += '******************\n'
        textout += \
            'Values for Ucfit\n'
        textout += 'a = {0: .5f} A, b = {1: .5f} A, c = {2: .5f} A\n'.\
            format(float(self.a), float(self.b), float(self.c))
        textout += 'alpha = {0: .5f}, beta = {1: .5f}, gamma = {2: .5f}\n'.\
            format(self.alpha, self.beta, self.gamma)
        textout += 'b/a = {0: .5f}, c/a = {1: .5f}\n'.\
            format(self.b / self.a, self.c / self.a)
        textout += 'V = {0: .5f} A^3\n\n'.format(float(self.v))
        textout += 'Below are the peaks at {0: 6.1f} GPa, {1: 5.0f} K\n'.\
            format(pressure, temperature)
        textout += 'd-spacing (A), intensity (%), h, k, l\n'
        for dl in self.DiffLines:
            textout += \
                "{0: .5f}, {1: .1f}, {2: .0f}, {3: .0f}, {4: .0f}\n".\
                format(float(dl.dsp), dl.intensity, dl.h, dl.k, dl.l)
        return textout


class JCPDS(object):
    """
    Class that defines a single JCPDS card.
    see __init__ for the involved parameters
    """

    def __init__(self, filename=None):
        self.phase_kind = 'jcpds'
        self.reference_pressure = 0.0
        self.reference_temperature = 300.0
        if filename is None:
            self.file = ' '
            self.name = ' '
            self.version = 0
            self.comments = []
            self.symmetry = ''
            self.k0 = 0.
            self.k0p = 0.  # k0p at 298K
            self.thermal_expansion = 0.  # alphat at 298K
            self.a0 = 0.
            self.b0 = 0.
            self.c0 = 0.
            self.alpha0 = 0.
            self.beta0 = 0.
            self.gamma0 = 0.
            self.v0 = 0.
            self.DiffLines = []
        else:
            self.file = filename
            self.read_file(self.file)
        self.a = 0.
        self.b = 0.
        self.c = 0.
        self.alpha = 0.
        self.beta = 0.
        self.gamma = 0.
        self.v = 0.
        self._cache_cal_dsp_key = None
        self._cache_tth = {}
        self._cache_hkl_key = None
        self._cache_hkl_text = None

    def _invalidate_plot_cache(self):
        self._cache_cal_dsp_key = None
        self._cache_tth = {}
        self._cache_hkl_key = None
        self._cache_hkl_text = None

    def read_file(self, file):
        """
        read a jcpds file
        """
        self.file = file
        self._invalidate_plot_cache()
        self.phase_kind = 'jcpds'
        # Construct base name = file without path and without extension
        name = os.path.splitext(os.path.basename(self.file))[0]
        self.name = name
        self.comments = []
        self.DiffLines = []
        if os.path.splitext(str(file))[1].lower() == '.rapo':
            self._read_rapo(file)
            return
        # Prefer PeakPo-style parser first; if it fails, fallback to Dioptas.
        try:
            self._read_peakpo_style(file)
        except Exception:
            self._read_dioptas_style(file)

        self._cal_v0()
        self.v = self.v0
        self.a = self.a0
        self.b = self.b0
        self.c = self.c0
        self.alpha = self.alpha0
        self.beta = self.beta0
        self.gamma = self.gamma0

    def _read_rapo(self, file):
        with open(file, 'r', encoding='utf-8') as f:
            payload = json.load(f)
        self.phase_kind = 'rapo'
        self.file = file
        self.name = str(payload.get('name') or os.path.splitext(os.path.basename(file))[0])
        self.version = int(payload.get('version', 1))
        self.comments = str(payload.get('comments', ''))
        self.symmetry = 'nosymmetry'
        self.k0 = 0.0
        self.k0p = 0.0
        self.thermal_expansion = 0.0
        self.a0 = 1.0
        self.b0 = 1.0
        self.c0 = 1.0
        self.alpha0 = 90.0
        self.beta0 = 90.0
        self.gamma0 = 90.0
        self.v0 = 1.0
        self.reference_pressure = float(payload.get('reference_pressure', 0.0))
        self.reference_temperature = float(payload.get('reference_temperature', 300.0))
        self.DiffLines = []
        for idx, mode in enumerate(payload.get('modes', [])):
            line = DiffractionLine()
            line.label = str(mode.get('name', f'mode_{idx + 1}'))
            line.dsp0 = float(mode.get('shift0', mode.get('shift', 0.0)))
            line.dsp = line.dsp0
            line.intensity = float(mode.get('intensity', 100.0))
            line.h = idx + 1
            line.pressure_coeffs = [
                float(val) for val in mode.get('pressure_coeffs', mode.get('pressure', []))
            ]
            line.temperature_coeffs = [
                float(val) for val in mode.get('temperature_coeffs', mode.get('temperature', []))
            ]
            self.DiffLines.append(line)
        self.v = self.v0
        self.a = self.a0
        self.b = self.b0
        self.c = self.c0
        self.alpha = self.alpha0
        self.beta = self.beta0
        self.gamma = self.gamma0

    def _make_cal_dsp_key(self, pressure, temperature, b_a, c_a, use_table_for_0GPa):
        return (
            float(pressure),
            float(temperature),
            None if b_a is None else float(b_a),
            None if c_a is None else float(c_a),
            bool(use_table_for_0GPa),
            str(self.symmetry),
            float(self.k0),
            float(self.k0p),
            float(self.v0),
            float(self.thermal_expansion),
            float(self.a0),
            float(self.b0),
            float(self.c0),
            float(self.alpha0),
            float(self.beta0),
            float(self.gamma0),
            len(self.DiffLines),
        )

    def _read_peakpo_style(self, file):
        inp = open(file, 'r').readlines()
        self.version = int(inp[0])  # JCPDS version number
        header = inp[1]  # header
        self.comments = header

        item = str.split(inp[2])
        crystal_system = int(item[0])
        if crystal_system == 1:
            self.symmetry = 'cubic'
        elif crystal_system == 2:
            self.symmetry = 'hexagonal'
        elif crystal_system == 3:
            self.symmetry = 'tetragonal'
        elif crystal_system == 4:
            self.symmetry = 'orthorhombic'
        elif crystal_system == 5:
            self.symmetry = 'monoclinic'
        elif crystal_system == 6:
            self.symmetry = 'triclinic'
        elif crystal_system == 7:
            self.symmetry = 'nosymmetry'
        # 1 cubic, 2 hexagonal, 3 tetragonal, 4 orthorhombic
        # 5 monoclinic, 6 triclinic, 7 no nosymmetry P, d-sp input

        self.k0 = float(item[1])
        self.k0p = float(item[2])

        item = str.split(inp[3])  # line for unit-cell parameters

        if crystal_system == 1:  # cubic
            a = float(item[0])
            b = a
            c = a
            alpha = 90.
            beta = 90.
            gamma = 90.
        elif crystal_system == 7:  # P, d-sp input
            a = float(item[0])
            b = a
            c = a
            alpha = 90.
            beta = 90.
            gamma = 90.
        elif crystal_system == 2:  # hexagonal
            a = float(item[0])
            c = float(item[1])
            b = a
            alpha = 90.
            beta = 90.
            gamma = 120.
        elif crystal_system == 3:  # tetragonal
            a = float(item[0])
            c = float(item[1])
            b = a
            alpha = 90.
            beta = 90.
            gamma = 90.
        elif crystal_system == 4:  # orthorhombic
            a = float(item[0])
            b = float(item[1])
            c = float(item[2])
            alpha = 90.
            beta = 90.
            gamma = 90.
        elif crystal_system == 5:  # monoclinic
            a = float(item[0])
            b = float(item[1])
            c = float(item[2])
            beta = float(item[3])
            alpha = 90.
            gamma = 90.
        elif crystal_system == 6:  # triclinic
            a = float(item[0])
            b = float(item[1])
            c = float(item[2])
            alpha = float(item[3])
            beta = float(item[4])
            gamma = float(item[5])

        self.a0 = a
        self.b0 = b
        self.c0 = c
        self.alpha0 = alpha
        self.beta0 = beta
        self.gamma0 = gamma

        item = str.split(inp[4])
        if self.version == 3:
            self.thermal_expansion = 0.
        else:
            self.thermal_expansion = float(item[0])

        for line in inp[6:]:
            item = str.split(line)
            if len(item) != 5:
                break
            diff_line = DiffractionLine()
            diff_line.dsp0 = float(item[0])
            diff_line.intensity = float(item[1])
            diff_line.h = float(item[2])
            diff_line.k = float(item[3])
            diff_line.l = float(item[4])
            self.DiffLines.append(diff_line)

    def _read_dioptas_style(self, file):
        # Parse Dioptas/JCPDSTools keyword format robustly:
        # "KEY: value" with flexible spaces/tabs.
        comments = []
        symmetry_upper = ''

        with open(file, 'r') as fp:
            for raw_line in fp:
                line = raw_line.strip()
                if (line == '') or line.startswith('#') or (':' not in line):
                    continue
                tag, value = line.split(':', 1)
                tag = tag.strip().upper()
                value = value.strip()

                if tag == 'VERSION':
                    try:
                        self.version = int(float(value))
                    except Exception:
                        self.version = 4
                elif tag == 'COMMENT':
                    comments.append(value)
                elif tag == 'K0':
                    self.k0 = float(value)
                elif tag == 'K0P':
                    self.k0p = float(value)
                elif tag == 'SYMMETRY':
                    symmetry_upper = value.upper()
                elif tag == 'A':
                    self.a0 = float(value)
                elif tag == 'B':
                    self.b0 = float(value)
                elif tag == 'C':
                    self.c0 = float(value)
                elif tag == 'ALPHA':
                    self.alpha0 = float(value)
                elif tag == 'BETA':
                    self.beta0 = float(value)
                elif tag == 'GAMMA':
                    self.gamma0 = float(value)
                elif tag == 'VOLUME':
                    self.v0 = float(value)
                elif tag == 'ALPHAT':
                    self.thermal_expansion = float(value)
                elif tag == 'DIHKL':
                    nums = re.findall(r'[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?', value)
                    if len(nums) < 5:
                        continue
                    diff_line = DiffractionLine()
                    diff_line.dsp0 = float(nums[0])
                    diff_line.dsp = diff_line.dsp0
                    diff_line.intensity = float(nums[1])
                    diff_line.h = float(nums[2])
                    diff_line.k = float(nums[3])
                    diff_line.l = float(nums[4])
                    self.DiffLines.append(diff_line)

        if self.version == 0:
            raise ValueError("Not a Dioptas-style JCPDS file")

        self.comments = comments[0] if len(comments) > 0 else ''
        self.symmetry = symmetry_upper.lower()

        # Fill symmetry-dependent defaults for missing cell parameters.
        if self.symmetry == 'cubic':
            self.b0 = self.a0
            self.c0 = self.a0
            self.alpha0 = 90.
            self.beta0 = 90.
            self.gamma0 = 90.
        elif self.symmetry == 'tetragonal':
            self.b0 = self.a0
            self.alpha0 = 90.
            self.beta0 = 90.
            self.gamma0 = 90.
        elif self.symmetry == 'orthorhombic':
            self.alpha0 = 90.
            self.beta0 = 90.
            self.gamma0 = 90.
        elif self.symmetry in ('hexagonal', 'trigonal'):
            self.b0 = self.a0
            self.alpha0 = 90.
            self.beta0 = 90.
            self.gamma0 = 120.
            # PeakPo uses hexagonal equations for trigonal.
            self.symmetry = 'hexagonal'
        elif self.symmetry == 'monoclinic':
            self.alpha0 = 90.
            self.gamma0 = 90.

        if self.v0 == 0.:
            self.v0 = cal_UnitCellVolume(
                self.symmetry, self.a0, self.b0, self.c0,
                self.alpha0, self.beta0, self.gamma0
            )

    def _cal_v0(self):
        """
        Computes the unit cell volume of the material at zero pressure and
        temperature from the unit cell parameters.
        """
        self.v0 = cal_UnitCellVolume(self.symmetry,
                                     self.a0, self.b0, self.c0,
                                     self.alpha0, self.beta0, self.gamma0)

    def _cal_v(self, pressure=0., temperature=300.):
        """
        Calculate volume at high pressure
        """
        pressure_st = pressure - self.thermal_expansion * \
            self.k0 * (temperature - 300.)
        if pressure == 0.0:  # p = 0 GPa
            self.v = self.v0
        else:
            if self.symmetry == 'nosymmetry':
                self.v = self.v0
            else:
                # print(pressure_st, self.v0, self.k0, self.k0p)
                try:
                    v_temp = bm3_v(pressure_st, self.v0, self.k0, self.k0p,
                                min_strain=0.3)
                except:
                    n_pts = 110
                    v_min = self.v0*0.1
                    v_calib = np.linspace(self.v0, v_min, n_pts)
                    p_calib = bm3_p(v_calib, self.v0, self.k0, self.k0p)
                    bm3s_v = interpolate.InterpolatedUnivariateSpline(p_calib, v_calib)
                    v_temp = bm3s_v(pressure) 
                    del_p = pressure - bm3_p(v_temp, self.v0, self.k0, self.k0p)
                    print(str(datetime.datetime.now())[:-7], 
                        ': Spline used for JCPDS calculation, residue = {0:.3e}'.format(del_p))
                self.v = v_temp

    def cal_dsp(self, pressure=0., temperature=300.,
                b_a=None, c_a=None, use_table_for_0GPa=True):
        """
        b_a and c_a are newly included for adjusting axial ratios.
        For cubic structure, these two inputs are ignored.
        For tetragonal and hexagonal, only c_a will be used.

        recalculate_zero = False: use the table d-spacing value for 0 GPa
        """

        cal_key = self._make_cal_dsp_key(
            pressure, temperature, b_a, c_a, use_table_for_0GPa)
        if cal_key == self._cache_cal_dsp_key:
            return
        if self.phase_kind == 'rapo':
            dp = float(pressure) - float(self.reference_pressure)
            dt = float(temperature) - float(self.reference_temperature)
            for dl in self.DiffLines:
                shift = float(dl.dsp0)
                for order, coeff in enumerate(getattr(dl, 'pressure_coeffs', []), start=1):
                    shift += float(coeff) * (dp ** order)
                for order, coeff in enumerate(getattr(dl, 'temperature_coeffs', []), start=1):
                    shift += float(coeff) * (dt ** order)
                dl.dsp = shift
            self._cache_cal_dsp_key = cal_key
            self._cache_tth = {}
            return
        # angles are always set to original values, so no need to reset
        # self.alpha = self.alpha0;
        # self.beta = self.beta0; self.gamma = self.gamma0
        self._cal_v(pressure, temperature)
        if b_a is None:
            b_a = self.b0 / self.a0
        if c_a is None:
            c_a = self.c0 / self.a0

        if (((pressure == 0.0) and (temperature == 300.) and (use_table_for_0GPa))
                or (self.symmetry == 'nosymmetry')):
            # p = 0 GPa, resetting to uc0 is necessary
            self.v = self.v0
            self.a = self.a0
            self.b = self.b0
            self.c = self.c0
            for dl in self.DiffLines:
                dsp = dl.dsp0
                dl.dsp = dsp
#            DLines = self.get_DiffractionLines()
#            for dl in DLines:
#                dsp = dl.dsp0
#                dl.dsp = dsp
        else:
            self._cal_UCPatPT(b_a, c_a)
            DLines = self.get_DiffractionLines()
#            a = self.a; b = self.b; c = self.c;
#            alpha = self.alpha; beta = self.beta; gamma = self.gamma
            for dl in DLines:
                dsp = cal_dspacing(self.symmetry, dl.h, dl.k, dl.l,
                                   self.a, self.b, self.c,
                                   self.alpha, self.beta, self.gamma)
                dl.dsp = dsp
            # the code worked without the following line, perhaps
            # due to the referencing or copy issue in python list
            self.DiffLines = DLines[:]
        """
        elif ((pressure == 0.0) and (temperature == 300.)) and (not use_table_for_0GPa):
            self._cal_UCPatPT(b_a, c_a)
            DLines = self.get_DiffractionLines()
#            a = self.a; b = self.b; c = self.c;
#            alpha = self.alpha; beta = self.beta; gamma = self.gamma
            for dl in DLines:
                dsp = cal_dspacing(self.symmetry, dl.h, dl.k, dl.l,
                                   self.a0, self.b0, self.c0,
                                   self.alpha0, self.beta0, self.gamma0)
                dl.dsp = dsp
            # the code worked without the following line, perhaps
            # due to the referencing or copy issue in python list
            self.DiffLines = DLines[:]
        """
        self._cache_cal_dsp_key = cal_key
        self._cache_tth = {}

    def get_DiffractionLines(self):
        """
        Returns the information for each reflection for the material.
        This information is an array of elements of class jcpds_reflection
        """
        return self.DiffLines

    def get_tthVSint(self, wavelength):
        """
        Returns twoth and intensity for bar plots in PyPeakPo
        Note that tth and intensity are numpy array not list
        If P, T, b_a, c_a have changed, run cal_dsp first for update
        """
#        self.cal_dsp(pressure, temperature, b_a, c_a)
        cache_key = (float(wavelength), self._cache_cal_dsp_key, len(self.DiffLines))
        cached = self._cache_tth.get(cache_key)
        if cached is not None:
            return cached
        DLines = self.get_DiffractionLines()
        if self.phase_kind == 'rapo':
            shift = np.fromiter((float(line.dsp) for line in DLines), dtype=float, count=len(DLines))
            intensity = np.fromiter((float(line.intensity) for line in DLines), dtype=float, count=len(DLines))
            out = (shift, intensity)
            self._cache_tth[cache_key] = out
            return out
        dsp = np.fromiter((float(line.dsp) for line in DLines), dtype=float, count=len(DLines))
        intensity = np.fromiter((float(line.intensity) for line in DLines), dtype=float, count=len(DLines))
        reciprocal_d = wavelength / 2. / dsp
        with np.errstate(invalid="ignore"):
            tth = 2. * np.degrees(np.arcsin(reciprocal_d))
        if np.any(reciprocal_d > 1.0):
            print(str(datetime.datetime.now())[:-7], 
                ": Ignore arcsin warning. Some dsp in jcpds are too small.")
        out = (tth, intensity)
        self._cache_tth[cache_key] = out
        return out

    def get_hkl_in_text(self):
        hkl_key = tuple((line.h, line.k, line.l) for line in self.DiffLines)
        if hkl_key == self._cache_hkl_key and self._cache_hkl_text is not None:
            return self._cache_hkl_text
        hkl = []
        for line in self.DiffLines:
            hkl.append("{0:.0f} {1:.0f} {2:.0f}".format(line.h, line.k, line.l))
        self._cache_hkl_key = hkl_key
        self._cache_hkl_text = hkl
        return hkl

    def find_DiffLine(self, tth_c, wavelength):
        """
        Returns index of the cloest diffraction line, its difference
            in abs tth
        If P, T, b_a, c_a have changed, run cal_dsp first for update
        """
        tth, intensity = self.get_tthVSint(wavelength)

        idx = (np.abs(tth - tth_c)).argmin()

        return idx, abs(tth[idx] - tth_c), tth[idx]

    def _cal_UCPatPT(self, b_a, c_a):
        if self.symmetry == 'cubic':
            self.a = (self.v)**(1. / 3.)
            self.b = self.a
            self.c = self.a
        elif (self.symmetry == 'hexagonal') or (self.symmetry == 'trigonal'):
            # self.a = (2. * self.v / (np.sqrt(3.)*self.c0/self.a0) )**(1./3.)
            self.a = (2. * self.v / (np.sqrt(3.) * c_a))**(1. / 3.)
            self.b = self.a
            self.c = self.a * c_a
        elif self.symmetry == 'tetragonal':
            # self.a = (self.v/(self.c0/self.a0))**(1./3.) ; self.b = self.a
            self.a = (self.v / (c_a))**(1. / 3.)
            self.b = self.a
            self.c = c_a * self.a
        elif self.symmetry == 'orthorhombic':
            # self.a = (self.v/(self.b0/self.a0*self.c0/self.a0))**(1./3.)
            self.a = (self.v / (b_a * c_a))**(1. / 3.)
            self.c = c_a * self.a
            self.b = b_a * self.a
        elif self.symmetry == 'monoclinic':
            # self.a = ( self.v / (self.b0/self.a0*self.c0/self.a0*\
            #        np.sin(np.radians(self.beta0))))**(1./3.)
            self.a = (self.v / (b_a * c_a *
                                np.sin(np.radians(self.beta0))))**(1. / 3.)
            self.c = c_a * self.a
            self.b = b_a * self.a
        elif self.symmetry == 'triclinic':
            a_term = np.sqrt(1. - (np.cos(np.radians(self.alpha0)))**2. -
                             (np.cos(np.radians(self.beta0)))**2. -
                             (np.cos(np.radians(self.gamma0)))**2. +
                             2. * np.cos(np.radians(self.alpha0)) *
                             np.cos(np.radians(self.beta0)) *
                             np.cos(np.radians(self.gamma0)))
            # self.a = (self.v/(self.b0/self.a0*self.c0/self.a0*a_term))
            # **(1./3.)
            self.a = (self.v / (b_a * c_a * a_term))**(1. / 3.)
            self.c = c_a * self.a
            self.b = b_a * self.a
        else:
            print(str(datetime.datetime.now())[:-7], 
                ': No symmetry is given')

#    return {'ver': ver, 'header': header, 'crystal_system': crystal_system, \
#            'K0': K0, 'K0p': K0p, 'u_param':
#               np.asarray([a,b,c,alpha,beta,gamma]), \
#            'alpha': thermal_expansion, 'peaks':
#           np.asarray([dsp, intensity, h, k, l]).T, \
#            'b_a': b_a, 'c_a': c_a}
    def set_from_cif(self, fn_cif, k0, k0p, file=None, name='', int_min=0.1,
                     version=4, comments='', thermal_expansion=0.,
                     two_theta_range=(0., 30.)):
        """
        define a JCPDS from CIF file.  made for jupyter notebook

        Parameters
        ----------
        fn_cif = file name and path of cif
        k0 = bulk modulus
        k0p = pressure derivative of k0
        file = name of file, optional (default = None)
        name = name of JCPDS, optional (default = '')
        version = version for JCPDS format, optional (default = 4)
        comments = comments for JCPDS file, optional (default = '')
        thermal_expansion = thermal expansion, optional (default = 0.)

        Returns
        -------

        """
        _require_pymatgen_for_cif()

        mg_version = mg.__version__
        mg_version_split = mg_version.split('.')
        d_current_version = datetime.datetime(int(mg_version_split[0]),
                                              int(mg_version_split[1]),
                                              int(mg_version_split[2]))
        d_work = datetime.datetime(2019, 4, 11)

        if d_current_version < d_work:
            print(str(datetime.datetime.now())[:-7], 
                ': Update your pymatgen newer than 2019.4.11.')
            return False

        structure = mg.Structure.from_file(fn_cif)
        crystal_system = SpacegroupAnalyzer(structure).get_crystal_system()
        if crystal_system == 'triclinic':
            print(str(datetime.datetime.now())[:-7], 
                ': Your cif has triclinic symmetry.  PeakPo cannot currently handle this.')
            return False
        self.set_from_pymatgen(structure, k0, k0p, file=file,
                               name=name, version=version,
                               comments=comments, int_min=int_min,
                               thermal_expansion=thermal_expansion)
        return True

    def set_from_pymatgen(self, structure, k0, k0p, file=None,
                          name='', version=4, int_min=0.1,
                          comments='', thermal_expansion=0.,
                          two_theta_range=(0., 30.)):
        """
        set parameters from pymatgen outputs

        Parameters
        ----------
        structure = pymatgen structure object
        k0 = bulk modulus
        k0p = pressure derivative of bulk modulus
        file = file name (optional)
        name = name (optional)
        version = version (optional)
        comments = comments (optional)
        thermal_expansion = 0 (optional)

        Returns
        -------

        """
        lattice = structure.lattice
        self.k0 = k0
        self.k0p = k0p
        self.a0 = float(lattice.a)
        self.b0 = float(lattice.b)
        self.c0 = float(lattice.c)
        self.alpha0 = float(lattice.alpha)
        self.beta0 = float(lattice.beta)
        self.gamma0 = float(lattice.gamma)
        symm = str(SpacegroupAnalyzer(structure).get_crystal_system())
        if symm == 'trigonal':
            self.symmetry = 'hexagonal'
        else:
            self.symmetry = symm
        self.file = file
        self.name = name
        self.version = version
        self.comments = comments
        self.thermal_expansion = thermal_expansion
        self.v0 = cal_UnitCellVolume(self.symmetry,
                                     self.a0, self.b0, self.c0,
                                     self.alpha0, self.beta0,
                                     self.gamma0)
        c = XRDCalculator(wavelength=0.3344)
        pattern = c.get_pattern(structure, two_theta_range=two_theta_range)
        h = []
        k = []
        l = []
        for i in range(pattern.hkls.__len__()):
            h.append(pattern.hkls[i][0]['hkl'][0])
            k.append(pattern.hkls[i][0]['hkl'][1])
            l.append(pattern.hkls[i][0]['hkl'][-1])
        d_lines = np.transpose([pattern.x, pattern.d_hkls, pattern.y, h, k, l])
        DiffLines = []
        for line in d_lines:
            if float(line[2]) >= int_min:
                d_line = DiffractionLine()
                d_line.dsp0 = float(line[1])
                #d_line.dsp = float(line[1])
                d_line.intensity = float(line[2])
                d_line.h = float(line[3])
                d_line.k = float(line[4])
                d_line.l = float(line[5])
                DiffLines.append(d_line)
        self.DiffLines = DiffLines

    def write_to_string(self, comments=" ",
                      calculate_1bar_table = False,
                      int_min=0., dsp_min=0.):

        textoutput = "{:d}\n".format(self.version)
        textoutput += comments + "\n"

        # 1 cubic, 2 hexagonal, 3 tetragonal, 4 orthorhombic
        # 5 monoclinic, 6 triclinic, 7 nosymmetry P, d-sp input

        str_el = "{0:.2f}   {1:.2f}".format(self.k0, self.k0p)

        if self.symmetry == 'cubic':  # cubic
            crystal_system = '1   '
            str_uc = "{0:.5f}".format(self.a0)
        elif self.symmetry == 'nosymmetry':  # P, d-sp input
            crystal_system = '7   '
            str_uc = "{0:.5f}".format(self.a0)
        elif self.symmetry == 'hexagonal' or self.symmetry == 'trigonal':
            crystal_system = '2   '
            str_uc = "{0:.5f}   {1:.5f}".format(self.a0, self.c0)
        elif self.symmetry == 'tetragonal':  # tetragonal
            crystal_system = '3   '
            str_uc = "{0:.5f}   {1:.5f}".format(self.a0, self.c0)
        elif self.symmetry == 'orthorhombic':  # orthorhombic
            crystal_system = '4   '
            str_uc = "{0:.5f}   {1:.5f}   {2:.5f}".format(self.a0,
                                                      self.b0, self.c0)
        elif self.symmetry == 'monoclinic':  # monoclinic
            crystal_system = '5   '
            str_uc = "{0:.5f}   {1:.5f}   {2:.5f}   {3:.5f}".format(
                self.a0, self.b0, self.c0, self.beta0)
        elif self.symmetry == 'triclinic':  # triclinic
            crystal_system = '6   '
            str_uc = "{0:.5f}   {1:.5f}   {2:.5f}   {3:.5f}   {4:.5f} \
                {5:.5f}".format(
                self.a0, self.b0, self.c0,
                self.alpha0, self.beta0, self.gamma0)

        textoutput += (crystal_system + str_el + " \n")
        textoutput += str_uc + " \n"
        textoutput += "{:.4e} \n".format(self.thermal_expansion)
        textoutput += "d-spacing    I/I0     h   k   l \n"

        if calculate_1bar_table:
            self.cal_dsp(use_table_for_0GPa=False)
            for line in self.DiffLines:
                if (line.dsp > dsp_min) and (line.intensity > int_min):
                    textoutput += \
                        "{0:.6f}   {1:.2f}   {2:.1f}   {3:.1f}   {4:.1f} \n".format(
                        line.dsp, line.intensity, line.h, line.k, line.l)
        else:
            for line in self.DiffLines:
                if (line.dsp0 > dsp_min) and (line.intensity > int_min):
                    textoutput += \
                        "{0:.6f}   {1:.2f}   {2:.1f}   {3:.1f}   {4:.1f} \n".format(
                        line.dsp0, line.intensity, line.h, line.k, line.l)
        return textoutput


    def write_to_file(self, filename, comments=" ",
                      calculate_1bar_table = False,
                      int_min=0., dsp_min=0.):
        """
        write a JCPDS file

        Parameters
        ----------
        filename = path and name of file

        Returns
        -------

        """
        textoutput = self.write_to_string(comments=comments,
                          calculate_1bar_table = calculate_1bar_table,
                          int_min=int_min, dsp_min=dsp_min)
        f = open(filename, 'w')
        f.write(textoutput)
        f.close()

    def write_to_dioptas_jcpds(self, filename, int_min=0., dsp_min=0.):
        jcpds_dioptas = jcpds()
        jcpds_dioptas.params['comments'].append(self.comments)
        jcpds_dioptas.params['symmetry'] = self.symmetry
        jcpds_dioptas.params['k0'] = self.k0
        jcpds_dioptas.params['k0p0'] = self.k0p
        jcpds_dioptas.params['alpha_t0'] = self.thermal_expansion
        jcpds_dioptas.params['a0'] = self.a0
        jcpds_dioptas.params['b0'] = self.b0
        jcpds_dioptas.params['c0'] = self.c0
        jcpds_dioptas.params['alpha0'] = self.alpha0
        jcpds_dioptas.params['beta0'] = self.beta0
        jcpds_dioptas.params['gamma0'] = self.gamma0
        jcpds_dioptas.params['v0'] = self.v0

        reflections = []

        for line in self.DiffLines:
            if (line.dsp0 > dsp_min) and (line.intensity > int_min):
                reflection = jcpds_reflection()
                reflection.d0 = line.dsp0
                reflection.intensity = line.intensity
                reflection.h = line.h
                reflection.k = line.k
                reflection.l = line.l
                reflections.append(reflection)
        jcpds_dioptas.reflections = reflections

        jcpds_dioptas.save_file(filename)


class Session(object):
    '''
    Developed for PeakPo and shared with PeakFt
    From 2015/3/15, PeakPo save using this class
    '''

    def __init__(self):
        self.pattern = None  # linked spectrum/pattern object
        self.waterfallpatterns = []  # list of pattern object
        self.wavelength = 0.0  # normal value at GSECARS
        self.pressure = 0.0  # DONOT CHANGE, matches with Qt Designer file
        self.temperature = 300.  # DONOT CHANGE, matches with Qt Designer file
        self.jlist = []  # list for JCPDS
        self.bg_roi = [4., 16.]
        self.bg_params = [20, 10, 20]
        self.jcpds_path = ''  # redundant, but easier to code
        self.chi_path = ''  # reduendant, but easier to code


class JCPDSplt(JCPDS):
    """
    Define new JPCDS for plot.  Add two attributes, color and show switch
    Developed for supporting PeakPo but used and shared with PeakFt
    """

    def __init__(self):
        ''' _org parameter needed : self.k0, self.k0p,
        self.thermal_expansion, self.v0
        '''
        super(JCPDSplt, self).__init__()
        self.color = ''
        self.display = True
        self.maxint = 1.0
        self.twk_b_a = 1.0
        self.twk_c_a = 1.0
        self.twk_v0 = 1.0
        self.twk_k0 = 1.0
        self.twk_k0p = 1.0
        self.twk_thermal_expansion = 1.0
        self.twk_int = 1.0

    def _sync_org_from_current(self):
        """Cache untweaked baseline values after any load/set operation."""
        self.k0_org = self.k0
        self.k0p_org = self.k0p
        self.v0_org = self.v0
        self.thermal_expansion_org = self.thermal_expansion

    def read_file(self, file):
        '''
        *_twk are tweaked parameters, and twk_* are tweaking coefficients
        a0, b0, c0, alpha, beta, gamma, v0 should not be tweaked
        '''
        super(JCPDSplt, self).read_file(file)
        self._sync_org_from_current()

    def set_from_cif(self, fn_cif, k0, k0p, file=None, name='', version=4,
                     comments='', int_min=0.1, thermal_expansion=0.):
        """
        Ensure CIF-created cards also initialize *_org tweak baselines.
        """
        success = super(JCPDSplt, self).set_from_cif(
            fn_cif, k0, k0p, file=file, name=name, version=version,
            comments=comments, int_min=int_min,
            thermal_expansion=thermal_expansion
        )
        if success:
            self._sync_org_from_current()
        return success

    def cal_dsp(self, pressure, temperature, b_a=None, c_a=None,
                use_table_for_0GPa=True):
        '''DiffLines are tweaked one unlike other notations'''
        if self.phase_kind == 'rapo':
            super(JCPDSplt, self).cal_dsp(
                pressure, temperature, b_a, c_a,
                use_table_for_0GPa=use_table_for_0GPa)
            return
        # set tweaked parameters
        self.k0 = self.k0_org * self.twk_k0
        self.k0p = self.k0p_org * self.twk_k0p
        self.v0 = self.v0_org * self.twk_v0
        self.thermal_expansion = self.thermal_expansion_org * \
            self.twk_thermal_expansion
        self.b_a = (self.b0 / self.a0) * self.twk_b_a
        self.c_a = (self.c0 / self.a0) * self.twk_c_a
        # get DiffLines
        super(JCPDSplt, self).cal_dsp(pressure, temperature,
                                      self.b_a, self.c_a,
                                      use_table_for_0GPa=use_table_for_0GPa)

    def get_dsp(self):
        dsp = []
        for dl in self.DiffLines:
            d = float(dl.dsp)
            dsp.append(d)
        return dsp

    def make_TextOutput(self, pressure, temperature):
        if self.phase_kind == 'rapo':
            textout = 'File from: ' + self.file + '\n'
            textout += 'Format: RAPO\n'
            textout += 'Comment from original file: ' + self.comments + '\n\n'
            textout += 'Mode, shift (cm-1), intensity\n'
            for dl in self.DiffLines:
                label = getattr(dl, 'label', 'mode')
                textout += ' {0}, {1:10.5f}, {2:10.1f}\n'.format(
                    label, float(dl.dsp), float(dl.intensity))
            return textout
        textout = 'File from: ' + self.file + '\n'
        textout += 'Version: ' + str(self.version) + '\n'
        textout += 'Comment from original file: ' + self.comments + '\n'
        textout += '\n'
        textout += 'Name: ' + self.name +'\n'
        textout += 'Crystal system: ' + self.symmetry + '\n'
        textout += '\n'
        textout += 'Values at high P-T after tweak \n'
        textout += ' a = {0:.5f} A, b = {1:.5f} A, c = {2:.5f} A\n'.\
            format(float(self.a), float(self.b), float(self.c))
        textout += ' alpha = {0:.2f}, beta = {1:.2f}, gamma = {2:.2f}\n'.\
            format(self.alpha, self.beta, self.gamma)
        textout += ' V = {0:.5f} A^3\n'.format(float(self.v))
        textout += '\n'
        if self.symmetry != 'nosymmetry':
            textout += 'Values at 1 bar and 300 K after tweak \n'
            a0_twk, b0_twk, c0_twk = get_cell_prm_twk(
                self.symmetry, self.v0, self.a0, self.b0, self.c0,
                self.alpha0, self.beta0, self.gamma0, self.twk_b_a, self.twk_c_a)
            textout += ' a0 = {0:.5f} A, b0 = {1:.5f} A, c0 = {2:.5f} A\n'.\
                format(float(a0_twk), float(b0_twk), float(c0_twk))
            textout += ' V0 = {0:.5f} A^3\n'.format(self.v0)  # v0 is tweaked value
            textout += ' K0 = {0:.1f} GPa, K0p = {1:.2f}, alpha = {2:.2e}\n'.\
                format(self.k0, self.k0p, self.thermal_expansion)
            textout += ' b/a = {0:.5f}, c/a = {1:.5f}\n'.\
                format(self.b / self.a, self.c / self.a)
            textout += ' Tweak for b/a = {0:.5f}, Tweak for c/a = {1:.5f}\n'.\
                format(self.twk_b_a, self.twk_c_a)
            textout += '\n'
            textout += 'Values in original JCPDS file\n'
            textout += ' a0 = {0:.5f} A, b0 = {1:.5f} A, c0 = {2:.5f} A\n'.\
                format(self.a0, self.b0, self.c0)
            textout += ' alpha0 = {0:.2f}, beta0 = {1:.2f}, gamma0 = {2:.2f}\n'.\
                format(self.alpha0, self.beta0, self.gamma0)
            textout += ' b0/a0 = {0:.5f}, c0/a0 = {1:.5f}\n'.\
                format(self.b0 / self.a0, self.c0 / self.a0)
            textout += ' V0 = {0:.5f} A^3\n'.format(self.v0_org)
            textout += ' K0 = {0:.1f} GPa, K0p = {1:.2f}, alpha = {2:.2e}\n'.\
                format(self.k0_org, self.k0p_org, self.thermal_expansion_org)
        textout += '\nBelow are the peaks at {0:6.2f} GPa, {1:5.0f} K\n'.\
            format(pressure, temperature)
        textout += ' d-spacing (A), intensity (%), h, k, l\n'
        for dl in self.DiffLines:
            textout += \
                " {0:10.5f}, {1:10.1f}, {2:5.0f}, {3:5.0f}, {4:5.0f}\n".\
                format(float(dl.dsp), dl.intensity, dl.h, dl.k, dl.l)

        return textout

    def write_to_twk_jcpds(self, filename, comments=" "):
        """
        write a twk JCPDS file

        Parameters
        ----------
        filename = path and name of file

        Returns
        -------

        """
        if self.phase_kind == 'rapo':
            payload = {
                'name': self.name,
                'version': self.version,
                'comments': comments,
                'reference_pressure': self.reference_pressure,
                'reference_temperature': self.reference_temperature,
                'modes': [],
            }
            for dl in self.DiffLines:
                payload['modes'].append({
                    'name': getattr(dl, 'label', 'mode'),
                    'shift0': float(dl.dsp0),
                    'intensity': float(dl.intensity),
                    'pressure_coeffs': list(getattr(dl, 'pressure_coeffs', [])),
                    'temperature_coeffs': list(getattr(dl, 'temperature_coeffs', [])),
                })
            with open(filename, 'w', encoding='utf-8') as f:
                json.dump(payload, f, indent=2)
            return
        f = open(filename, 'w')
        f.write("{:d}\n".format(self.version))
        f.write(comments + "\n")

        # 1 cubic, 2 hexagonal, 3 tetragonal, 4 orthorhombic
        # 5 monoclinic, 6 triclinic, 7 nosymmetry P, d-sp input

        str_el = "{0:.2f} {1:.2f}".format(self.k0, self.k0p)

        a0_twk, b0_twk, c0_twk = get_cell_prm_twk(
            self.symmetry, self.v0, self.a0, self.b0, self.c0,
            self.alpha0, self.beta0, self.gamma0, self.twk_b_a, self.twk_c_a)

        if self.symmetry == 'cubic':  # cubic
            crystal_system = '1 '
            str_uc = "{0:.5f}".format(a0_twk)
        elif self.symmetry == 'nosymmetry':  # P, d-sp input
            crystal_system = '7 '
            str_uc = "{0:.5f}".format(a0_twk)
        elif self.symmetry == 'hexagonal' or self.symmetry == 'trigonal':
            crystal_system = '2 '
            str_uc = "{0:.5f} {1:.5f}".format(a0_twk, c0_twk)
        elif self.symmetry == 'tetragonal':  # tetragonal
            crystal_system = '3 '
            str_uc = "{0:.5f} {1:.5f}".format(a0_twk, c0_twk)
        elif self.symmetry == 'orthorhombic':  # orthorhombic
            crystal_system = '4 '
            str_uc = "{0:.5f} {1:.5f} {2:.5f}".format(a0_twk,
                                                      b0_twk, c0_twk)
        elif self.symmetry == 'monoclinic':  # monoclinic
            crystal_system = '5 '
            str_uc = "{0:.5f} {1:.5f} {2:.5f} {3:.5f}".format(
                a0_twk, b0_twk, c0_twk, self.beta0)
        elif self.symmetry == 'triclinic':  # triclinic
            crystal_system = '6 '
            str_uc = "{0:.5f} {1:.5f} {2:.5f} {3:.5f} {4:.5f} \
                {5:.5f}".format(
                a0_twk, b0_twk, c0_twk,
                self.alpha0, self.beta0, self.gamma0)

        f.write(crystal_system + str_el + " \n")
        f.write(str_uc + " \n")
        f.write("{:.4e} \n".format(self.thermal_expansion))
        f.write("d-spacing    I/I0     h   k   l \n")

        self.cal_dsp(0., 300., use_table_for_0GPa=False)
        for line in self.DiffLines:
            f.write("{0:.6f} {1:.2f} {2:.1f} {3:.1f} {4:.1f} \n".format(
                line.dsp0, line.intensity, line.h, line.k, line.l))
        f.close()


def get_cell_prm_twk(symmetry, v_twk, a0, b0, c0, alpha0, beta0, gamma0,
                     twk_b_a, twk_c_a):
    b_a_twk = b0 / a0 * twk_b_a
    c_a_twk = c0 / a0 * twk_c_a
    if symmetry == 'cubic':
        a_twk = (v_twk)**(1. / 3.)
        b_twk = a_twk
        c_twk = a_twk
    elif (symmetry == 'hexagonal') or (symmetry == 'trigonal'):
        a_twk = (2. * v_twk / (np.sqrt(3.) * c_a_twk))**(1. / 3.)
        b_twk = a_twk
        c_twk = a_twk * c_a_twk
    elif symmetry == 'tetragonal':
        a_twk = (v_twk / (c_a_twk))**(1. / 3.)
        b_twk = a_twk
        c_twk = c_a_twk * a_twk
    elif symmetry == 'orthorhombic':
        a_twk = (v_twk / (b_a_twk * c_a_twk))**(1. / 3.)
        c_twk = c_a_twk * a_twk
        b_twk = b_a_twk * a_twk
    elif symmetry == 'monoclinic':
        a_twk = (v_twk / (b_a_twk * c_a_twk *
                          np.sin(np.radians(beta0))))**(1. / 3.)
        c_twk = c_a_twk * a_twk
        b_twk = b_a_twk * a_twk
    elif symmetry == 'triclinic':
        a_term = np.sqrt(1. - (np.cos(np.radians(alpha0)))**2. -
                         (np.cos(np.radians(beta0)))**2. -
                         (np.cos(np.radians(gamma0)))**2. +
                         2. * np.cos(np.radians(alpha0)) *
                         np.cos(np.radians(beta0)) *
                         np.cos(np.radians(gamma0)))
        a_twk = (v_twk / (b_a_twk * c_a_twk * a_term))**(1. / 3.)
        c_twk = c_a_twk * a_twk
        b_twk = b_a_twk * a_twk
    else:
        #print('no symmetry is given')
        a_twk = a0
        b_twk = b0
        c_twk = c0

    return a_twk, b_twk, c_twk



from qtpy import QtWidgets
from qtpy import QtCore
from qtpy import QtGui
from ..utils import SpinBoxFixStyle
from .mplcontroller import MplController


class JcpdsTableController(object):
    COL_SHOW = 0
    COL_LOCK = 1
    COL_COLOR = 2
    COL_V0 = 3
    COL_BA = 4
    COL_CA = 5
    COL_INT = 6
    COL_K0 = 7
    COL_K0P = 8
    COL_ALPHA = 9

    def __init__(self, model, widget):
        self.model = model
        self.widget = widget
        self.plot_ctrl = MplController(self.model, self.widget)

    def _apply_changes_to_graph(self, limits=None):
        self.plot_ctrl.update(limits=limits)

    def sync_model_from_table(self):
        """
        Commit any in-progress table spinbox edits to model values before save.
        """
        table = self.widget.tableWidget_JCPDS
        n_rows = min(table.rowCount(), len(self.model.jcpds_lst))
        for row in range(n_rows):
            phase = self.model.jcpds_lst[row]
            item0 = table.item(row, self.COL_SHOW)
            if item0 is not None:
                phase.display = (item0.checkState() == QtCore.Qt.Checked)
            item1 = table.item(row, self.COL_LOCK)
            if item1 is not None:
                phase._pkpo_locked = (item1.checkState() == QtCore.Qt.Checked)

            def _commit_spin(col, attr):
                box = table.cellWidget(row, col)
                if box is None:
                    return
                try:
                    box.interpretText()
                except Exception:
                    pass
                setattr(phase, attr, float(box.value()))

            _commit_spin(self.COL_V0, "twk_v0")
            _commit_spin(self.COL_BA, "twk_b_a")
            _commit_spin(self.COL_CA, "twk_c_a")
            _commit_spin(self.COL_INT, "twk_int")
            _commit_spin(self.COL_K0, "twk_k0")
            _commit_spin(self.COL_K0P, "twk_k0p")
            _commit_spin(self.COL_ALPHA, "twk_thermal_expansion")

    def update(self, step=0.001):
        """
        show jcpds cards in the QTableWidget
        """
        n_columns = self.COL_ALPHA + 1
        n_rows = self.model.jcpds_lst.__len__()
        self.widget.tableWidget_JCPDS.setColumnCount(n_columns)
        self.widget.tableWidget_JCPDS.setRowCount(n_rows)
        self.widget.tableWidget_JCPDS.horizontalHeader().setVisible(True)
        self.widget.tableWidget_JCPDS.verticalHeader().setVisible(True)
        self.widget.tableWidget_JCPDS.setHorizontalHeaderLabels(
            ['👁',      # 0 - show checkbox
             '🔒',      # 1 - lock checkbox
             '',        # 2 - color (clickable if unlocked)
             'V0 twk',  # 3
             'b/a twk',  # 4
             'c/a twk',  # 5
             'Int twk',  # 6
             'K0 twk',  # 7
             'K0p twk',  # 8
             'alpha0 twk'])  # 9
        self.widget.tableWidget_JCPDS.setVerticalHeaderLabels(
            [phase.name for phase in self.model.jcpds_lst])
        
        self.widget.tableWidget_JCPDS.setColumnWidth(self.COL_SHOW, 30)
        self.widget.tableWidget_JCPDS.setColumnWidth(self.COL_LOCK, 45)
        self.widget.tableWidget_JCPDS.setColumnWidth(self.COL_COLOR, 35)
        
        for row in range(n_rows):
            phase = self.model.jcpds_lst[row]
            row_locked = bool(getattr(phase, "_pkpo_locked", False))

            # column 0 - show checkbox
            item0 = QtWidgets.QTableWidgetItem()
            item0.setFlags(QtCore.Qt.ItemIsUserCheckable |
                           QtCore.Qt.ItemIsEnabled)
            if phase.display:
                item0.setCheckState(QtCore.Qt.Checked)
            else:
                item0.setCheckState(QtCore.Qt.Unchecked)
            self.widget.tableWidget_JCPDS.setItem(row, self.COL_SHOW, item0)

            # column 1 - lock checkbox
            item_lock = QtWidgets.QTableWidgetItem()
            item_lock.setFlags(QtCore.Qt.ItemIsUserCheckable |
                               QtCore.Qt.ItemIsEnabled)
            item_lock.setCheckState(
                QtCore.Qt.Checked if row_locked else QtCore.Qt.Unchecked)
            self.widget.tableWidget_JCPDS.setItem(row, self.COL_LOCK, item_lock)
            
            # column 2 - color (clickable only when unlocked)
            item_color = QtWidgets.QTableWidgetItem('')
            item_color.setBackground(QtGui.QColor(phase.color))
            item_color.setFlags(QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable)
            self.widget.tableWidget_JCPDS.setItem(row, self.COL_COLOR, item_color)
            
            # column 3 - V0 tweak
            if (phase.symmetry == 'nosymmetry'):
                item_v0 = QtWidgets.QTableWidgetItem(' ')
                item_v0.setFlags(QtCore.Qt.ItemIsEnabled)
                self.widget.tableWidget_JCPDS.setItem(row, self.COL_V0, item_v0)
            else:
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk = \
                    QtWidgets.QDoubleSpinBox()
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.setAlignment(
                    QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                    QtCore.Qt.AlignVCenter)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.setMaximum(2.0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.setSingleStep(
                    step)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.setDecimals(3)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.setProperty(
                    "value", phase.twk_v0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.valueChanged.\
                    connect(self._handle_doubleSpinBoxChanged)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.setStyle(
                    SpinBoxFixStyle())
                self.widget.tableWidget_JCPDS.setCellWidget(
                    row, self.COL_V0, self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.setFocusPolicy(
                    QtCore.Qt.StrongFocus)
                self.widget.tableWidget_JCPDS_doubleSpinBox_V0twk.\
                    setKeyboardTracking(False)
            
            # column 4 - b/a tweak
            if (phase.symmetry == 'cubic') or \
                    (phase.symmetry == 'tetragonal') or \
                    (phase.symmetry == 'hexagonal') or \
                    (phase.symmetry == 'nosymmetry'):
                item_b_a = QtWidgets.QTableWidgetItem(' ')
                item_b_a.setFlags(QtCore.Qt.ItemIsEnabled)
                self.widget.tableWidget_JCPDS.setItem(row, self.COL_BA, item_b_a)
            else:
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk = \
                    QtWidgets.QDoubleSpinBox()
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.\
                    setAlignment(
                        QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                        QtCore.Qt.AlignVCenter)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.setMaximum(
                    2.0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.\
                    setSingleStep(step)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.setDecimals(
                    3)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.setProperty(
                    "value", phase.twk_b_a)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.\
                    valueChanged.connect(
                        self._handle_doubleSpinBoxChanged)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.setStyle(
                    SpinBoxFixStyle())
                self.widget.tableWidget_JCPDS.setCellWidget(
                    row, self.COL_BA, self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.\
                    setKeyboardTracking(False)
                self.widget.tableWidget_JCPDS_doubleSpinBox_b_atwk.\
                    setFocusPolicy(QtCore.Qt.StrongFocus)
            
            # column 5 - c/a tweak
            if (phase.symmetry == 'cubic') or \
                    (phase.symmetry == 'nosymmetry'):
                item_c_a = QtWidgets.QTableWidgetItem(' ')
                item_c_a.setFlags(QtCore.Qt.ItemIsEnabled)
                self.widget.tableWidget_JCPDS.setItem(row, self.COL_CA, item_c_a)
            else:
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk = \
                    QtWidgets.QDoubleSpinBox()
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.\
                    setAlignment(
                        QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                        QtCore.Qt.AlignVCenter)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.setMaximum(
                    2.0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.\
                    setSingleStep(step)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.setDecimals(
                    3)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.setProperty(
                    "value", phase.twk_c_a)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.\
                    valueChanged.connect(
                        self._handle_doubleSpinBoxChanged)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.setStyle(
                    SpinBoxFixStyle())
                self.widget.tableWidget_JCPDS.setCellWidget(
                    row, self.COL_CA, self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.\
                    setKeyboardTracking(False)
                self.widget.tableWidget_JCPDS_doubleSpinBox_c_atwk.\
                    setFocusPolicy(QtCore.Qt.StrongFocus)
            
            # column 6 - Int tweak
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk = \
                QtWidgets.QDoubleSpinBox()
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.setAlignment(
                QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                QtCore.Qt.AlignVCenter)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.setMaximum(1.0)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.setSingleStep(
                step)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.setDecimals(3)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.setProperty(
                "value", phase.twk_int)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.valueChanged.\
                connect(self._handle_doubleSpinBoxChanged)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.setStyle(
                SpinBoxFixStyle())
            self.widget.tableWidget_JCPDS.setCellWidget(
                row, self.COL_INT, self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.setFocusPolicy(
                QtCore.Qt.StrongFocus)
            self.widget.tableWidget_JCPDS_doubleSpinBox_inttwk.\
                setKeyboardTracking(False)
            
            # column 7 - K0 tweak
            if (phase.symmetry == 'nosymmetry'):
                item_k0 = QtWidgets.QTableWidgetItem(' ')
                item_k0.setFlags(QtCore.Qt.ItemIsEnabled)
                self.widget.tableWidget_JCPDS.setItem(row, self.COL_K0, item_k0)
            else:
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk = \
                    QtWidgets.QDoubleSpinBox()
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.setAlignment(
                    QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                    QtCore.Qt.AlignVCenter)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.setMaximum(
                    2.0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.setSingleStep(
                    step)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.setDecimals(3)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.setProperty(
                    "value", phase.twk_k0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.valueChanged.\
                    connect(self._handle_doubleSpinBoxChanged)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.setStyle(
                    SpinBoxFixStyle())
                self.widget.tableWidget_JCPDS.setCellWidget(
                    row, self.COL_K0, self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.setFocusPolicy(
                    QtCore.Qt.StrongFocus)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0twk.\
                    setKeyboardTracking(False)
            
            # column 8 - K0p tweak
            if (phase.symmetry == 'nosymmetry'):
                item_k0p = QtWidgets.QTableWidgetItem(' ')
                item_k0p.setFlags(QtCore.Qt.ItemIsEnabled)
                self.widget.tableWidget_JCPDS.setItem(row, self.COL_K0P, item_k0p)
            else:
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk = \
                    QtWidgets.QDoubleSpinBox()
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.setAlignment(
                    QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                    QtCore.Qt.AlignVCenter)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.setMaximum(2.0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.setSingleStep(
                    step)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.setDecimals(3)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.setProperty(
                    "value", phase.twk_k0p)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.valueChanged.\
                    connect(self._handle_doubleSpinBoxChanged)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.setStyle(
                    SpinBoxFixStyle())
                self.widget.tableWidget_JCPDS.setCellWidget(
                    row, self.COL_K0P, self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.setFocusPolicy(
                    QtCore.Qt.StrongFocus)
                self.widget.tableWidget_JCPDS_doubleSpinBox_K0ptwk.\
                    setKeyboardTracking(False)
            
            # column 9 - alpha0 tweak
            if (phase.symmetry == 'nosymmetry'):
                item_alpha0 = QtWidgets.QTableWidgetItem(' ')
                item_alpha0.setFlags(QtCore.Qt.ItemIsEnabled)
                self.widget.tableWidget_JCPDS.setItem(row, self.COL_ALPHA, item_alpha0)
            else:
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk = \
                    QtWidgets.QDoubleSpinBox()
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.setAlignment(
                    QtCore.Qt.AlignRight | QtCore.Qt.AlignTrailing |
                    QtCore.Qt.AlignVCenter)
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.setMaximum(
                    2.0)
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.\
                    setSingleStep(step)
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.setDecimals(
                    3)
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.setProperty(
                    "value", phase.twk_thermal_expansion)
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.\
                    valueChanged.connect(self._handle_doubleSpinBoxChanged)
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.setStyle(
                    SpinBoxFixStyle())
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.\
                    setFocusPolicy(QtCore.Qt.StrongFocus)
                self.widget.tableWidget_JCPDS.setCellWidget(
                    row, self.COL_ALPHA, self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk)
                self.widget.tableWidget_JCPDS_doubleSpinBox_alpha0twk.\
                    setKeyboardTracking(False)
            self._set_row_locked_state(row, row_locked)
        
        self.widget.tableWidget_JCPDS.resizeColumnsToContents()
        # Disconnect all previous connections to avoid multiple dialogs
        try:
            self.widget.tableWidget_JCPDS.itemClicked.disconnect()
        except:
            pass  # No connections exist yet
        self.widget.tableWidget_JCPDS.itemClicked.connect(
            self._handle_ItemClicked)

    def _handle_doubleSpinBoxChanged(self, value):
        box = self.widget.sender()
        index = self.widget.tableWidget_JCPDS.indexAt(box.pos())
        if index.isValid():
            idx = index.row()
            if index.column() == self.COL_V0:
                self.model.jcpds_lst[idx].twk_v0 = value
            elif index.column() == self.COL_BA:
                self.model.jcpds_lst[idx].twk_b_a = value
            elif index.column() == self.COL_CA:
                self.model.jcpds_lst[idx].twk_c_a = value
            elif index.column() == self.COL_INT:
                self.model.jcpds_lst[idx].twk_int = value
            elif index.column() == self.COL_K0:
                self.model.jcpds_lst[idx].twk_k0 = value
            elif index.column() == self.COL_K0P:
                self.model.jcpds_lst[idx].twk_k0p = value
            elif index.column() == self.COL_ALPHA:
                self.model.jcpds_lst[idx].twk_thermal_expansion = value
            
            if self.model.jcpds_lst[idx].display:
                self._apply_changes_to_graph()

    def _handle_ItemClicked(self, item):
        idx = item.row()
        
        # Handle show checkbox clicks (column 0)
        if item.column() == self.COL_SHOW:
            if (item.checkState() == QtCore.Qt.Checked) ==\
                    self.model.jcpds_lst[idx].display:
                return
            if item.checkState() == QtCore.Qt.Checked:
                self.model.jcpds_lst[idx].display = True
            elif item.checkState() == QtCore.Qt.Unchecked:
                self.model.jcpds_lst[idx].display = False
            self._apply_changes_to_graph()
        elif item.column() == self.COL_LOCK:
            is_locked = (item.checkState() == QtCore.Qt.Checked)
            self.model.jcpds_lst[idx]._pkpo_locked = bool(is_locked)
            self._set_row_locked_state(idx, is_locked)
        
        # Handle color clicks
        elif item.column() == self.COL_COLOR:
            if bool(getattr(self.model.jcpds_lst[idx], "_pkpo_locked", False)):
                return
            color = QtWidgets.QColorDialog.getColor(
                QtGui.QColor(self.model.jcpds_lst[idx].color))
            if color.isValid():
                item.setBackground(color)
                self.model.jcpds_lst[idx].color = str(color.name())
                self._apply_changes_to_graph()

    def _set_row_locked_state(self, row, locked):
        table = self.widget.tableWidget_JCPDS
        color_item = table.item(row, self.COL_COLOR)
        if color_item is not None:
            if locked:
                color_item.setFlags(QtCore.Qt.NoItemFlags)
            else:
                color_item.setFlags(QtCore.Qt.ItemIsEnabled |
                                    QtCore.Qt.ItemIsSelectable)
        for col in range(self.COL_V0, self.COL_ALPHA + 1):
            box = table.cellWidget(row, col)
            if box is not None:
                box.setEnabled(not locked)
            item = table.item(row, col)
            if item is not None:
                if locked:
                    item.setFlags(QtCore.Qt.NoItemFlags)
                else:
                    item.setFlags(QtCore.Qt.ItemIsEnabled)

    def update_steps_only(self, step):
        """
        show jcpds cards in the QTableWidget
        """
        self.widget.tableWidget_JCPDS.clear()
        self.update(step=step)



import os
import glob
import numpy as np

from ..utils import get_temp_dir, readchi
from ..ds_ramspec import SpectrumPattern


def load_spectrum_xy(spectrum_path, spectrum_cache):
    if spectrum_path in spectrum_cache:
        return spectrum_cache[spectrum_path]
    ext = os.path.splitext(str(spectrum_path))[1].lower()
    if ext == ".chi":
        __, __, x, y = readchi(spectrum_path)
        x = np.asarray(x, dtype=float)
        y = np.asarray(y, dtype=float)
    else:
        spectrum = SpectrumPattern()
        spectrum.read_file(spectrum_path)
        x, y = spectrum.get_raw()
        x = np.asarray(x, dtype=float)
        y = np.asarray(y, dtype=float)
    spectrum_cache[spectrum_path] = (x, y)
    return x, y


def load_bgsub_or_raw_xy(spectrum_path, use_bgsub, spectrum_cache):
    if not bool(use_bgsub):
        return load_spectrum_xy(spectrum_path, spectrum_cache)

    # Priority 1: temp bgsub file under <data>-rampo/
    try:
        temp_dir = get_temp_dir(spectrum_path)
        base = os.path.splitext(os.path.basename(spectrum_path))[0]
        temp_bgsub = os.path.join(temp_dir, f"{base}.bgsub.chi")
        if os.path.exists(temp_bgsub):
            __, __, x, y = readchi(temp_bgsub)
            return np.asarray(x, dtype=float), np.asarray(y, dtype=float)
    except Exception:
        pass

    # Priority 2: sibling bgsub file next to chi
    sibling_bgsub = os.path.splitext(spectrum_path)[0] + ".bgsub.chi"
    if os.path.exists(sibling_bgsub):
        __, __, x, y = readchi(sibling_bgsub)
        return np.asarray(x, dtype=float), np.asarray(y, dtype=float)

    # Fallback: raw chi if bgsub file is unavailable.
    return load_spectrum_xy(spectrum_path, spectrum_cache)


def find_temp_cake_triplet(chi_path):
    temp_dir = get_temp_dir(chi_path)
    tth_files = sorted(glob.glob(os.path.join(temp_dir, "*.tth.cake.npy")))
    if not tth_files:
        return None

    stem_map = {}
    for tth_f in tth_files:
        stem = tth_f[: -len(".tth.cake.npy")]
        azi_f = stem + ".azi.cake.npy"
        int_f = stem + ".int.cake.npy"
        if os.path.exists(azi_f) and os.path.exists(int_f):
            stem_map[stem] = (tth_f, azi_f, int_f)
    if not stem_map:
        return None

    # Most recent triplet first.
    triplets = sorted(stem_map.values(), key=lambda t: os.path.getmtime(t[2]), reverse=True)
    return triplets[0]


def load_cake_data(chi_path, cake_cache):
    if chi_path in cake_cache:
        return cake_cache[chi_path]

    triplet = find_temp_cake_triplet(chi_path)
    if triplet is None:
        return None

    tth = np.load(triplet[0])
    azi = np.load(triplet[1])
    intensity = np.load(triplet[2])
    payload = (
        np.asarray(tth, dtype=float),
        np.asarray(azi, dtype=float),
        np.asarray(intensity, dtype=float),
    )
    cake_cache[chi_path] = payload
    return payload


load_chi_xy = load_spectrum_xy



import numpy as np


def fit_bg_poly(x, y_obs, poly_order=5, fit_areas=None):
    x_arr = np.asarray(x, dtype=float).reshape(-1)
    y_arr = np.asarray(y_obs, dtype=float).reshape(-1)
    n = min(x_arr.size, y_arr.size)
    if n == 0:
        return y_arr[:0]
    x_arr = x_arr[:n]
    y_arr = y_arr[:n]
    valid = np.isfinite(x_arr) & np.isfinite(y_arr)
    if np.count_nonzero(valid) < 2:
        return np.asarray(y_arr, dtype=float)
    if fit_areas:
        area_mask = np.zeros_like(valid, dtype=bool)
        for area in fit_areas:
            try:
                xmin = float(area[0])
                xmax = float(area[1])
            except Exception:
                continue
            if xmax < xmin:
                xmin, xmax = xmax, xmin
            area_mask |= (x_arr >= xmin) & (x_arr <= xmax)
        fit_mask = valid & area_mask
        if np.count_nonzero(fit_mask) < 2:
            fit_mask = valid
    else:
        fit_mask = valid
    x_fit = x_arr[fit_mask]
    y_fit = y_arr[fit_mask]
    deg = max(0, min(int(poly_order), x_fit.size - 1))
    coeffs = np.polyfit(x_fit, y_fit, deg)
    y_bg = np.polyval(coeffs, x_arr)
    return np.asarray(y_bg, dtype=float)



import os


_REQUIRED_QT_API = "pyqt6"


def enforce_pyqt6():
    current = os.environ.get("QT_API")
    if current and current.lower() != _REQUIRED_QT_API:
        raise RuntimeError(
            f"RamPo requires QT_API={_REQUIRED_QT_API}, got {current!r}."
        )
    os.environ["QT_API"] = _REQUIRED_QT_API




from qtpy import QtGui, QtWidgets, QtCore
import os.path
from .fileutils import extract_extension


def dialog_savefile(obj, default_filename):
    """
    :return: "" if the user choose not to overwrite or save
    """
    extension = extract_extension(default_filename)
    extension_to_search = "(*." + extension + ")"
    reply = QtWidgets.QMessageBox.question(
        obj, 'Question',
        'Do you want to save in default filename, %s ?' % default_filename,
        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        QtWidgets.QMessageBox.Yes)
    if reply == QtWidgets.QMessageBox.No:
        new_filename = QtWidgets.QFileDialog.getSaveFileName(
            obj, "Choose different filename.",
            default_filename, extension_to_search)[0]
        return str(new_filename)
    else:
        if os.path.exists(default_filename):
            reply = QtWidgets.QMessageBox.question(
                obj, 'Question',
                'The file already exist.  Do you want to overwrite?',
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No)
            if reply == QtWidgets.QMessageBox.No:
                return ''
            else:
                return default_filename
        else:
            return default_filename


class ErrorMessageBox(QtWidgets.QDialog):
    """
    If possible merge with InformationBox below
    """

    def __init__(self, *args, **kwargs):
        super(ErrorMessageBox, self).__init__(*args, **kwargs)
        self.setWindowTitle("Error report")

        self.text_lbl = QtWidgets.QLabel()
        self.text_lbl.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.scroll_area = QtWidgets.QScrollArea()

        self.scroll_area.setWidget(self.text_lbl)
        self.scroll_area.setWidgetResizable(True)
        self.ok_btn = QtWidgets.QPushButton('OK')

        _layout = QtWidgets.QGridLayout()
        _layout.addWidget(self.scroll_area, 0, 0, 1, 10)
        _layout.addWidget(self.ok_btn, 1, 9)

        self.setLayout(_layout)
        self.ok_btn.clicked.connect(self.close)

    def setText(self, text_str):
        self.text_lbl.setText(text_str)


class InformationBox(QtWidgets.QDialog):
    def __init__(self, title="Information", *args, **kwargs):
        super(InformationBox, self).__init__(*args, **kwargs)
        self.setWindowTitle(title)

        self.text_lbl = QtWidgets.QLabel()
        self.text_lbl.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self.scroll_area = QtWidgets.QScrollArea()

        self.scroll_area.setWidget(self.text_lbl)
        self.scroll_area.setWidgetResizable(True)
        self.ok_btn = QtWidgets.QPushButton('OK')

        _layout = QtWidgets.QGridLayout()
        _layout.addWidget(self.scroll_area, 0, 0, 1, 10)
        _layout.addWidget(self.ok_btn, 1, 9)

        self.setLayout(_layout)
        self.ok_btn.clicked.connect(self.close)

    def setText(self, text_str):
        self.text_lbl.setText(text_str)



# -*- coding: utf-8 -*-
# T-Rax - GUI program for analysis of spectroscopy data during
# diamond anvil cell experiments
# Copyright (C) 2016 Clemens Prescher (clemens.prescher@gmail.com)
# Institute for Geology and Mineralogy, University of Cologne
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implements the SPE_File class for loading princeton instrument binary SPE files into Python
works for version 2 and version 3 files.

Usage:
mydata = SPE_File('data.spe')

most important properties:

num_frames - number of frames collected
exposure_time

img - 2d data if num_frames==1
      list of 2d data if num_frames>1

x_calibration - wavelength information of x-axis



the data will be automatically loaded and all important parameters and the data
can be requested from the object.
"""

import datetime
from xml.dom.minidom import parseString

import numpy as np
from numpy.polynomial.polynomial import polyval
from dateutil import parser


class SpeFile(object):
    def __init__(self, filename, debug=False):
        """Opens the PI SPE file and loads its content

        :param filename: filename of the PI SPE to open
        :param debug: if set to true, will automatically save <filename>.xml files for version 3 spe files in the spe
        directory
        """
        """"""
        self.filename = filename
        self.debug = debug
        self._fid = open(filename, "rb")
        self.version = self._get_version()
        self._read_parameter()
        self._read_img()
        self._fid.close()

    def _get_version(self):
        self.xml_offset = self._read_at(678, 1, np.int64)[0]
        if self.xml_offset == 0:
            return 2
        else:
            return 3

    def _read_parameter(self):
        """Reads in size and datatype. Decides whether it should check in the binary
        header (version 2) or in the xml-footer for the experimental parameters"""
        self._read_size()
        self._read_datatype()
        if self.version == 2:
            self._read_parameter_from_header()
        elif self.version == 3:
            try:
                self._read_parameter_from_dom()
            except ValueError:
                print("Error reading SPE file: %s" % self.filename)
                print("Trying to read parameters from header")
                self._read_parameter_from_header()
        else:
            raise ValueError("Unknown SPE file version: %s" % self.version)

    def _read_size(self):
        """reads the dimensions of the Model from the header into the object
        resulting object parameters are _xdim and _ydim"""
        self._xdim = np.int64(self._read_at(42, 1, np.int16)[0])
        self._ydim = np.int64(self._read_at(656, 1, np.int16)[0])

    def _read_parameter_from_header(self):
        """High level function which calls all the read_"parameter" function
        that are reading information from the binary header.
        """
        self._read_date_time_from_header()
        self._read_calibration_from_header()
        self._read_exposure_from_header()
        self._read_detector_from_header()
        self._read_grating_from_header()
        self._read_center_wavelength_from_header()
        self._read_roi_from_header()
        self._read_num_frames_from_header()
        self._read_num_combined_frames_from_header()

    def _read_parameter_from_dom(self):
        """High level function which calls all the read_"parameter" function
        that are reading information from the xml footer.
        """
        self._get_xml_string()
        self._create_dom_from_xml()
        self._read_date_time_from_dom()
        self._read_calibration_from_dom()
        self._read_detector_from_dom()
        self._read_exposure_from_dom()
        self._read_grating_from_dom()
        self._read_center_wavelength_from_dom()
        self._read_roi_from_dom()
        self._select_wavelength_from_roi()
        self._read_num_frames_from_header()
        self._read_num_combined_frames_from_dom()

    def _read_date_time_from_header(self):
        """Reads the collection time from the header into the date_time field"""
        rawdate = self._read_at(20, 9, np.int8)
        rawtime = self._read_at(172, 6, np.int8)
        strdate = "".join([chr(i) for i in rawdate])
        strdate += "".join([chr(i) for i in rawtime])
        import locale

        locale.setlocale(locale.LC_TIME, "en_US.utf8")
        try:
            self.date_time = datetime.datetime.strptime(str(strdate), "%d%b%Y%H%M%S")
        except:
            print("WARNING: could note read datetime from SPE_FILE")
            self.date_time = datetime.datetime.now()

    def _read_calibration_from_header(self):
        """Reads the calibration from the header into the x_calibration field"""
        x_polynocoeff = self._read_at(3263, 6, np.double)
        x_val = np.arange(self._xdim) + 1
        self.x_calibration = np.array(polyval(x_val, x_polynocoeff))

    def _read_exposure_from_header(self):
        """Reads the exposure time from the header into the exposure_time field"""
        self.exposure_time = self._read_at(10, 1, np.float32)
        self.exposure_time = self.exposure_time[0]

    def _read_detector_from_header(self):
        """Sets the detector value to unspecified, because the detector is not
        specified in the binary header. Only in the xml footer of version 3 SPE
        files"""
        self.detector = "unspecified"

    def _read_grating_from_header(self):
        """Reads grating position from the header into the grating field"""
        self.grating = str(self._read_at(650, 1, np.float32)[0])

    def _read_center_wavelength_from_header(self):
        """Reads center wavelength position from the header into the center_wavelength field"""
        self.center_wavelength = float(self._read_at(72, 1, np.float32)[0])

    def _read_roi_from_header(self):
        return

    def _read_num_frames_from_header(self):
        self.num_frames = self._read_at(1446, 1, np.int32)[0]

    def _read_num_combined_frames_from_header(self):
        self._num_combined_frames = 1

    def _create_dom_from_xml(self):
        """Creates a DOM representation of the xml footer and saves it in the
        dom field"""
        try:
            self.dom = parseString(self.xml_string)
        except:
            raise ValueError(
                "Could not parse XML footer of SPE file: %s" % self.filename
            )

    def _get_xml_string(self):
        """Reads out the xml string from the file end"""
        self._fid.seek(int(self.xml_offset))
        self.xml_string = self._fid.read()

        if self.debug:
            fid = open(self.filename + ".xml", "w")
            for line in self.xml_string:
                fid.write(line)
            fid.close()

    def _read_date_time_from_dom(self):
        """Reads the time of collection and saves it date_time field"""
        date_time_str = self.dom.getElementsByTagName("Origin")[0].getAttribute(
            "created"
        )
        self.date_time = parser.parse(date_time_str)

    def _read_calibration_from_dom(self):
        """Reads the x calibration of the image from the xml footer and saves
        it in the x_calibration field"""
        spe_format = self.dom.childNodes[0]
        calibrations = spe_format.getElementsByTagName("Calibrations")[0]
        wavelengthmapping = calibrations.getElementsByTagName("WavelengthMapping")[0]
        wavelengths = wavelengthmapping.getElementsByTagName("Wavelength")[0]
        wavelength_values = wavelengths.childNodes[0]
        self.x_calibration = np.array(
            [float(i) for i in wavelength_values.toxml().split(",")]
        )

    def _read_exposure_from_dom(self):
        """Reads th exposure time of the experiment into the exposure_time field"""
        if (
            len(self.dom.getElementsByTagName("Experiment")) != 1
        ):  # check if it is a real v3.0 file
            if (
                len(self.dom.getElementsByTagName("ShutterTiming")) == 1
            ):  # check if it is a pixis detector
                self._exposure_time = self.dom.getElementsByTagName("ExposureTime")[
                    0
                ].childNodes[0]
                self.exposure_time = float(self._exposure_time.toxml()) / 1000.0
            else:
                # self._exposure_time = self.dom.getElementsByTagName('ReadoutControl')[0]. \
                #     getElementsByTagName('Time')[0].childNodes[0].nodeValue
                # self._exposure_time = float(self._exposure_time)/1000000000
                self._exposure_time = (
                    self.dom.getElementsByTagName("Gating")[0]
                    .getElementsByTagName("RepetitiveGate")[0]
                    .getElementsByTagName("Pulse")[0]
                    .getAttribute("width")
                )
                self._exposure_time = float(self._exposure_time) / 1000000000
                self._accumulations = (
                    self.dom.getElementsByTagName("Accumulations")[0]
                    .childNodes[0]
                    .nodeValue
                )
                self.exposure_time = float(self._exposure_time) * float(
                    self._accumulations
                )
        else:  # this is searching for legacy experiment:
            self._exposure_time = (
                self.dom.getElementsByTagName("LegacyExperiment")[0]
                .getElementsByTagName("Experiment")[0]
                .getElementsByTagName("CollectionParameters")[0]
                .getElementsByTagName("Exposure")[0]
                .attributes["value"]
                .value
            )
            self.exposure_time = float(self._exposure_time.split()[0])

    def _read_detector_from_dom(self):
        """Reads the detector information from the dom object"""
        self._camera = self.dom.getElementsByTagName("Camera")
        if len(self._camera) >= 1:
            self.detector = self._camera[0].getAttribute("model")
        else:
            self.detector = "unspecified"

    def _read_grating_from_dom(self):
        """Reads the type of grating from the dom Model"""
        try:
            self._grating = (
                self.dom.getElementsByTagName("Devices")[0]
                .getElementsByTagName("Spectrometer")[0]
                .getElementsByTagName("Grating")[0]
                .getElementsByTagName("Selected")[0]
                .childNodes[0]
                .toxml()
            )
            self.grating = self._grating.split("[")[1].split("]")[0].replace(",", " ")
        except IndexError:
            self._read_grating_from_header()

    def _read_center_wavelength_from_dom(self):
        """Reads the center wavelength from the dom Model and saves it center_wavelength field"""
        try:
            self._center_wavelength = (
                self.dom.getElementsByTagName("Devices")[0]
                .getElementsByTagName("Spectrometer")[0]
                .getElementsByTagName("Grating")[0]
                .getElementsByTagName("CenterWavelength")[0]
                .childNodes[0]
                .toxml()
            )
            self.center_wavelength = float(self._center_wavelength)
        except IndexError:
            self._read_center_wavelength_from_header()

    def _read_roi_from_dom(self):
        """Reads the ROIs information defined in the SPE file.
        Depending on the modus it will read out:
        For CustomRegions
        roi_x, roi_y, roi_width, roi_height, roi_x_binning, roi_y_binning
        For FullSensor
        roi_x,roi_y, roi_width, roi_height"""
        try:
            self.roi_modus = str(
                self.dom.getElementsByTagName("ReadoutControl")[0]
                .getElementsByTagName("RegionsOfInterest")[0]
                .getElementsByTagName("Selection")[0]
                .childNodes[0]
                .toxml()
            )
            if self.roi_modus == "CustomRegions":
                self.roi_dom = (
                    self.dom.getElementsByTagName("ReadoutControl")[0]
                    .getElementsByTagName("RegionsOfInterest")[0]
                    .getElementsByTagName("CustomRegions")[0]
                    .getElementsByTagName("RegionOfInterest")[0]
                )
                self.roi_dom_width = (
                    self.dom.getElementsByTagName("DataFormat")[0]
                    .getElementsByTagName("DataBlock")[0]
                    .getElementsByTagName("DataBlock")[0]
                )
                self.roi_x = int(self.roi_dom.attributes["x"].value)
                self.roi_y = int(self.roi_dom.attributes["y"].value)
                self.roi_width = int(self.roi_dom_width.attributes["width"].value)
                self.roi_height = int(self.roi_dom.attributes["height"].value)
                self.roi_x_binning = int(self.roi_dom.attributes["xBinning"].value)
                self.roi_y_binning = int(self.roi_dom.attributes["yBinning"].value)
            elif self.roi_modus == "FullSensor":
                self.roi_x = 0
                self.roi_y = 0
                self.roi_width = self._xdim
                self.roi_height = self._ydim
            elif self.roi_modus == "LineSensor":
                self.roi_dom_width = (
                    self.dom.getElementsByTagName("DataFormat")[0]
                    .getElementsByTagName("DataBlock")[0]
                    .getElementsByTagName("DataBlock")[0]
                )
                self.roi_width = int(self.roi_dom_width.attributes["width"].value)
                self.roi_height = 1
                self.roi_x = 0
                self.roi_y = 0
                self._xdim = self.roi_width
                self._ydim = 1

        except IndexError:
            self.roi_x = 0
            self.roi_y = 0
            self.roi_width = self._xdim
            self.roi_height = self._ydim

    def _read_num_combined_frames_from_dom(self):
        try:
            self.frame_combination = (
                self.dom.getElementsByTagName("Experiment")[0]
                .getElementsByTagName("Devices")[0]
                .getElementsByTagName("Cameras")[0]
                .getElementsByTagName("FrameCombination")[0]
            )
            self.num_frames_combined = int(
                self.frame_combination.getElementsByTagName("FramesCombined")[0]
                .childNodes[0]
                .toxml()
            )
        except IndexError:
            self._read_num_combined_frames_from_header()

    def _select_wavelength_from_roi(self):
        try:
            self.x_calibration = self.x_calibration[
                self.roi_x : self.roi_x + self.roi_width
            ]
        except AttributeError:
            print("SPE File bad!")

    def _read_datatype(self):
        self._data_type = self._read_at(108, 1, np.uint16)[0]

    def _read_at(self, pos, size, ntype):
        pos = int(pos)
        size = int(size)
        self._fid.seek(pos)
        return np.fromfile(self._fid, ntype, size)

    def _read_img(self):
        self.img = self._read_frame(4100)
        if self.num_frames > 1:
            img_temp = []
            img_temp.append(self.img)
            for n in range(self.num_frames - 1):
                img_temp.append(self._read_frame())
            self.img = img_temp

    def _read_frame(self, pos=None):
        """Reads in a frame at a specific binary position. The following parameters have to
        be predefined before calling this function:
        datatype - either 0,1,2,3,8 for float32, int32, int16, uint16 or uint32 (datatypes can be found on page 10 in
        the SPE 3.0 File format manual)
        _xdim, _ydim - being the dimensions.
        """
        if pos == None:
            pos = self._fid.tell()
        if self._data_type == 0:
            img = self._read_at(pos, self._xdim * self._ydim, np.float32)
        elif self._data_type == 1:
            img = self._read_at(pos, self._xdim * self._ydim, np.int32)
        elif self._data_type == 2:
            img = self._read_at(pos, self._xdim * self._ydim, np.int16)
        elif self._data_type == 3:
            img = self._read_at(pos, self._xdim * self._ydim, np.uint16)
        elif self._data_type == 8:
            img = self._read_at(pos, self._xdim * self._ydim, np.uint32)
        return img.reshape((self._ydim, self._xdim))

    def get_index_from(self, wavelength):
        """
        calculating image index for a given index
        :param wavelength: wavelength in nm
        :return: index
        """
        result = []
        xdata = self.x_calibration
        try:
            for w in wavelength:
                try:
                    base_ind = max(max(np.where(xdata <= w)))
                    if base_ind < len(xdata) - 1:
                        result.append(
                            int(
                                np.round(
                                    (w - xdata[base_ind])
                                    / (xdata[base_ind + 1] - xdata[base_ind])
                                    + base_ind
                                )
                            )
                        )
                    else:
                        result.append(base_ind)
                except:
                    result.append(0)
            return np.array(result)
        except TypeError:
            base_ind = max(max(np.where(xdata <= wavelength)))
            return int(
                np.round(
                    (wavelength - xdata[base_ind])
                    / (xdata[base_ind + 1] - xdata[base_ind])
                    + base_ind
                )
            )

    def get_wavelength_from(self, index):
        if isinstance(index, list):
            result = []
            for c in index:
                result.append(self.x_calibration[c])
            return np.array(result)
        else:
            return self.x_calibration[index]

    def get_dimension(self):
        """Returns (xdim, ydim)"""
        return (self._xdim, self._ydim)

    def get_roi(self):
        """Returns the ROI which was defined by WinSpec or Lightfield for datacollection"""
        return [
            self.roi_x,
            self.roi_x + self.roi_width - 1,
            self.roi_y,
            self.roi_y + self.roi_height - 1,
        ]

    def get_file_size(self):
        self._fid.seek(0, 2)
        self.file_size = self._fid.tell()
        return self.file_size



import numpy as np
import scipy.constants as const

def convert_wl_to_energy(wavelength):
    """
    wavelength in anstrom
    """
    return np.ceil(const.speed_of_light * const.Planck /
           const.electron_volt / (wavelength * 1.e-10)) * 1.e-3



import json
import os
import shutil
import tempfile
import datetime
import io
import hashlib
import re
from dataclasses import dataclass
from typing import Optional

import numpy as np

from ..utils import get_temp_dir
from ..ds_ramspec import PatternPeakPo
from ..ds_jcpds import JCPDSplt, DiffractionLine
from ..ds_section import Section
from ..ds_ccd import DiffImg


MANIFEST_FILE = "rampo_manifest.json"
SESSION_FILE = "rampo_session.json"
SECTIONS_FILE = "rampo_sections.json"
UI_STATE_FILE = "rampo_ui_state.json"
JCPDS_FILE = "rampo_rapo.json"
BACKUP_INDEX_FILE = "rampo_backup_index.json"

FORMAT_FAMILY = "rampo-session"
FORMAT_VERSION = 1


@dataclass
class SaveResult:
    param_dir: str
    manifest_path: str
    backup_id: Optional[str]
    changed_files: list[str]


class _ParamLite:
    def __init__(self, value=None, stderr=None, vary=True):
        self.value = value
        self.stderr = stderr
        self.vary = vary


class _FitResultLite:
    def __init__(self, payload=None):
        payload = payload or {}
        self.chisqr = payload.get("chisqr")
        self.redchi = payload.get("redchi")
        self.aic = payload.get("aic")
        self.bic = payload.get("bic")
        self.best_fit = np.asarray(payload.get("best_fit", []), dtype=float)
        self._components = {
            k: np.asarray(v, dtype=float)
            for k, v in payload.get("components", {}).items()
        }
        self.params = {}
        for key, val in payload.get("params", {}).items():
            self.params[key] = _ParamLite(
                value=val.get("value"),
                stderr=val.get("stderr"),
                vary=val.get("vary", True),
            )

    def eval_components(self, x=None):
        return self._components


def _json_default(obj):
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, np.integer):
        return int(obj)
    if isinstance(obj, np.floating):
        return float(obj)
    if isinstance(obj, np.bool_):
        return bool(obj)
    if hasattr(obj, "__dict__"):
        return obj.__dict__
    return str(obj)


def _atomic_write_bytes(path, payload):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    tmp_path = None
    try:
        with tempfile.NamedTemporaryFile("wb", delete=False, dir=os.path.dirname(path)) as f:
            tmp_path = f.name
            f.write(payload)
            f.flush()
            os.fsync(f.fileno())
        os.replace(tmp_path, path)
    finally:
        if tmp_path and os.path.exists(tmp_path):
            os.remove(tmp_path)


def _atomic_write_json(path, data):
    payload = json.dumps(data, indent=2, sort_keys=True, default=_json_default).encode("utf-8")
    _atomic_write_bytes(path, payload)


def _file_bytes(path):
    with open(path, "rb") as f:
        return f.read()


def _file_differs_from_payload(path, payload, chunk_size=1024 * 1024):
    if not os.path.exists(path):
        return True
    try:
        if os.path.getsize(path) != len(payload):
            return True
        with open(path, "rb") as f:
            offset = 0
            payload_len = len(payload)
            while offset < payload_len:
                end = min(offset + chunk_size, payload_len)
                chunk = f.read(end - offset)
                if chunk != payload[offset:end]:
                    return True
                offset = end
        return False
    except Exception:
        return True


def _relpath_or_abs(path, root):
    if path is None:
        return None
    path = str(path)
    try:
        rel = os.path.relpath(path, root)
    except Exception:
        return path
    if rel.startswith(".."):
        return path
    return rel


def _resolve_path(stored_path, root):
    if stored_path is None:
        return None
    if os.path.isabs(stored_path):
        return stored_path
    return os.path.normpath(os.path.join(root, stored_path))


def _bg_temp_names(pattern):
    base = os.path.splitext(os.path.basename(pattern.fname))[0]
    return f"{base}.bgsub.chi", f"{base}.bg.chi"


def _processed_temp_name(pattern):
    base = os.path.splitext(os.path.basename(pattern.fname))[0]
    return f"{base}.chi"


def _smoothed_temp_name():
    return "smooth.chi"


def _serialize_pattern(
        pattern, chi_root, fallback_relpath=None, force_abs_fname=False):
    if pattern is None:
        return None
    bgsub_name, bg_name = _bg_temp_names(pattern)
    stored_fname = getattr(pattern, "_pkpo_original_fname", None) or getattr(pattern, "fname", None)
    if fallback_relpath is None:
        fallback_relpath = getattr(pattern, "_pkpo_fallback_relpath", None)
    stored_fname_out = str(stored_fname) if force_abs_fname else _relpath_or_abs(stored_fname, chi_root)
    return {
        "fname": stored_fname_out,
        "fallback_fname": fallback_relpath,
        "wavelength": getattr(pattern, "wavelength", None),
        "color": getattr(pattern, "color", None),
        "display": getattr(pattern, "display", None),
        "bg_roi": list(getattr(pattern, "roi", []) or []),
        "bg_params": list(getattr(pattern, "params_chbg", []) or []),
        "bg_fit_areas": list(getattr(pattern, "bg_fit_areas", []) or []),
        "bgsub_file": bgsub_name,
        "bg_file": bg_name,
    }


def _load_pattern(payload, chi_root, temp_dir):
    if payload is None:
        return None
    ptn = PatternPeakPo()
    fname_primary = _resolve_path(payload.get("fname"), chi_root)
    fallback_stored = payload.get("fallback_fname")
    if fallback_stored is None:
        fname_fallback = None
    elif os.path.isabs(str(fallback_stored)):
        fname_fallback = str(fallback_stored)
    else:
        fname_fallback = os.path.normpath(os.path.join(temp_dir, str(fallback_stored)))

    used_fallback = False
    if fname_primary and os.path.exists(fname_primary):
        ptn.read_file(fname_primary)
        ptn.fname = fname_primary
    elif fname_fallback and os.path.exists(fname_fallback):
        ptn.read_file(fname_fallback)
        ptn.fname = fname_fallback
        used_fallback = True
    else:
        ptn.fname = fname_primary
        ptn.x_raw = None
        ptn.y_raw = None
    ptn.wavelength = payload.get("wavelength", 0.3344)
    if hasattr(ptn, "apply_excitation_wavelength"):
        try:
            ptn.apply_excitation_wavelength(ptn.wavelength)
        except Exception:
            pass
    ptn.color = payload.get("color", "white")
    ptn.display = payload.get("display", False)
    ptn._pkpo_original_fname = fname_primary
    ptn._pkpo_fallback_fname = fname_fallback
    ptn._pkpo_fallback_relpath = fallback_stored
    ptn._pkpo_fallback_in_use = used_fallback
    # Keep background-subtracted state in legacy chi files.
    payload_fit_areas = payload.get("bg_fit_areas", []) or []
    payload_bg_params = payload.get("bg_params", []) or []
    payload_bg_roi = payload.get("bg_roi", []) or []
    if ptn.fname is not None:
        ptn.read_bg_from_tempfile(temp_dir=temp_dir)
    if payload_fit_areas:
        ptn.bg_fit_areas = [
            [float(area[0]), float(area[1])]
            for area in payload_fit_areas
            if isinstance(area, (list, tuple)) and len(area) >= 2
        ]
    if payload_bg_params:
        try:
            ptn.params_chbg = [int(payload_bg_params[0])]
        except Exception:
            pass
    if payload_bg_roi and len(payload_bg_roi) >= 2:
        try:
            ptn.roi = [float(payload_bg_roi[0]), float(payload_bg_roi[1])]
        except Exception:
            pass
    return ptn


def _serialize_jcpds_item(phase, chi_root):
    dlines = []
    for line in getattr(phase, "DiffLines", []):
        dlines.append(
            {
                "dsp": getattr(line, "dsp", 0.0),
                "dsp0": getattr(line, "dsp0", 0.0),
                "intensity": getattr(line, "intensity", 0.0),
                "h": getattr(line, "h", 0),
                "k": getattr(line, "k", 0),
                "l": getattr(line, "l", 0),
            }
        )
    return {
        "file": _relpath_or_abs(getattr(phase, "file", None), chi_root),
        "name": getattr(phase, "name", ""),
        "version": getattr(phase, "version", 0),
        "comments": getattr(phase, "comments", ""),
        "symmetry": getattr(phase, "symmetry", ""),
        "k0": getattr(phase, "k0", 0.0),
        "k0p": getattr(phase, "k0p", 0.0),
        "thermal_expansion": getattr(phase, "thermal_expansion", 0.0),
        "a0": getattr(phase, "a0", 0.0),
        "b0": getattr(phase, "b0", 0.0),
        "c0": getattr(phase, "c0", 0.0),
        "alpha0": getattr(phase, "alpha0", 0.0),
        "beta0": getattr(phase, "beta0", 0.0),
        "gamma0": getattr(phase, "gamma0", 0.0),
        "v0": getattr(phase, "v0", 0.0),
        "a": getattr(phase, "a", 0.0),
        "b": getattr(phase, "b", 0.0),
        "c": getattr(phase, "c", 0.0),
        "alpha": getattr(phase, "alpha", 0.0),
        "beta": getattr(phase, "beta", 0.0),
        "gamma": getattr(phase, "gamma", 0.0),
        "v": getattr(phase, "v", 0.0),
        "color": getattr(phase, "color", ""),
        "display": getattr(phase, "display", True),
        "maxint": getattr(phase, "maxint", 1.0),
        "twk_b_a": getattr(phase, "twk_b_a", 1.0),
        "twk_c_a": getattr(phase, "twk_c_a", 1.0),
        "twk_v0": getattr(phase, "twk_v0", 1.0),
        "twk_k0": getattr(phase, "twk_k0", 1.0),
        "twk_k0p": getattr(phase, "twk_k0p", 1.0),
        "twk_thermal_expansion": getattr(phase, "twk_thermal_expansion", 1.0),
        "twk_int": getattr(phase, "twk_int", 1.0),
        "locked": bool(getattr(phase, "_pkpo_locked", False)),
        "k0_org": getattr(phase, "k0_org", None),
        "k0p_org": getattr(phase, "k0p_org", None),
        "v0_org": getattr(phase, "v0_org", None),
        "thermal_expansion_org": getattr(phase, "thermal_expansion_org", None),
        "DiffLines": dlines,
    }


def _load_jcpds_item(payload, chi_root):
    phase = JCPDSplt()
    for k, v in payload.items():
        if k == "DiffLines":
            continue
        if k == "file":
            setattr(phase, k, _resolve_path(v, chi_root))
        else:
            setattr(phase, k, v)
    dlines = []
    for row in payload.get("DiffLines", []):
        line = DiffractionLine()
        line.dsp0 = row.get("dsp0", 0.0)
        line.dsp = row.get("dsp", 0.0)
        line.intensity = row.get("intensity", 0.0)
        line.h = row.get("h", 0)
        line.k = row.get("k", 0)
        line.l = row.get("l", 0)
        dlines.append(line)
    phase.DiffLines = dlines
    if getattr(phase, "k0_org", None) is None:
        phase.k0_org = getattr(phase, "k0", 0.0)
    if getattr(phase, "k0p_org", None) is None:
        phase.k0p_org = getattr(phase, "k0p", 0.0)
    if getattr(phase, "v0_org", None) is None:
        phase.v0_org = getattr(phase, "v0", 0.0)
    if getattr(phase, "thermal_expansion_org", None) is None:
        phase.thermal_expansion_org = getattr(phase, "thermal_expansion", 0.0)
    phase._pkpo_locked = bool(payload.get("locked", False))
    return phase


def _fit_result_to_dict(section):
    fit = getattr(section, "fit_result", None)
    if fit is None:
        return None
    out = {
        "chisqr": getattr(fit, "chisqr", None),
        "redchi": getattr(fit, "redchi", None),
        "aic": getattr(fit, "aic", None),
        "bic": getattr(fit, "bic", None),
        "params": {},
        "best_fit": None,
        "components": {},
    }
    params = getattr(fit, "params", {}) or {}
    for key, prm in params.items():
        out["params"][key] = {
            "value": getattr(prm, "value", None),
            "stderr": getattr(prm, "stderr", None),
            "vary": getattr(prm, "vary", True),
        }
    best_fit = getattr(fit, "best_fit", None)
    if best_fit is not None:
        out["best_fit"] = np.asarray(best_fit).tolist()
    try:
        comps = fit.eval_components(x=section.x)
        out["components"] = {k: np.asarray(v).tolist() for k, v in comps.items()}
    except Exception:
        out["components"] = {}
    return out


def _sanitize_name(name):
    return re.sub(r"[^A-Za-z0-9_.-]", "_", str(name))


def _trim_trailing_nan(arr):
    arr = np.asarray(arr, dtype=float).reshape(-1)
    if arr.size == 0:
        return arr
    valid = np.where(~np.isnan(arr))[0]
    if valid.size == 0:
        return np.asarray([], dtype=float)
    return arr[: valid[-1] + 1]


def _component_sort_key(name):
    key = str(name).strip().lower()
    m = re.fullmatch(r"p(\d+)", key)
    if m is not None:
        return (0, int(m.group(1)))
    if key == "b":
        return (2, 0)
    return (1, key)


def _compute_section_csv_payload(columns):
    if not columns:
        return b""
    col_names = [name for name, _ in columns]
    n_rows = max(np.asarray(values, dtype=float).reshape(-1).size for _, values in columns)
    matrix = np.full((n_rows, len(columns)), np.nan, dtype=float)
    for idx, (_, values) in enumerate(columns):
        vals = np.asarray(values, dtype=float).reshape(-1)
        matrix[:vals.size, idx] = vals
    buf = io.StringIO()
    np.savetxt(
        buf,
        matrix,
        fmt="%.18g",
        delimiter=",",
        header=",".join(col_names),
        comments="",
    )
    return buf.getvalue().encode("utf-8")


def _load_section_csv_columns(path):
    try:
        if (path is None) or (not os.path.exists(path)):
            return {}
        data = np.genfromtxt(path, delimiter=",", names=True, dtype=float, encoding="utf-8")
        if data is None:
            return {}
        names = getattr(getattr(data, "dtype", None), "names", None)
        if not names:
            return {}
        out = {}
        for name in names:
            col = data[name]
            arr = np.asarray([float(col)], dtype=float) if np.isscalar(col) else np.asarray(col, dtype=float)
            out[name] = _trim_trailing_nan(arr)
        return out
    except Exception:
        return {}


def _load_csv_array_legacy(path):
    try:
        if (path is None) or (not os.path.exists(path)):
            return np.asarray([], dtype=float)
        with open(path, "r", encoding="utf-8") as f:
            text = f.read().strip()
        if text == "":
            return np.asarray([], dtype=float)
        return np.loadtxt(io.StringIO(text), delimiter=",", dtype=float, ndmin=1)
    except Exception:
        return np.asarray([], dtype=float)


def _fit_result_to_dict_with_csv(section):
    fit = getattr(section, "fit_result", None)
    if fit is None:
        return None
    out = {
        "chisqr": getattr(fit, "chisqr", None),
        "redchi": getattr(fit, "redchi", None),
        "aic": getattr(fit, "aic", None),
        "bic": getattr(fit, "bic", None),
        "params": {},
        "best_fit_column": None,
        "component_columns": {},
    }
    params = getattr(fit, "params", {}) or {}
    for key, prm in params.items():
        out["params"][key] = {
            "value": getattr(prm, "value", None),
            "stderr": getattr(prm, "stderr", None),
            "vary": getattr(prm, "vary", True),
        }
    return out


def _section_to_dict(section, section_tag, section_payloads):
    columns = []
    if section.x is not None:
        columns.append(("x", section.x))
    if section.y_bgsub is not None:
        columns.append(("y_bgsub", section.y_bgsub))
    if section.y_bg is not None:
        columns.append(("y_bg", section.y_bg))

    fit_payload = _fit_result_to_dict_with_csv(section)
    fit = getattr(section, "fit_result", None)
    if fit is not None:
        best_fit = getattr(fit, "best_fit", None)
        if best_fit is not None:
            columns.append(("fit_best", best_fit))
            fit_payload["best_fit_column"] = "fit_best"
        try:
            comps = fit.eval_components(x=section.x) or {}
        except Exception:
            comps = {}
        for comp_name in sorted(comps.keys(), key=_component_sort_key):
            col_name = f"comp_{_sanitize_name(comp_name)}"
            columns.append((col_name, comps[comp_name]))
            fit_payload["component_columns"][comp_name] = col_name

    section_csv_file = None
    section_csv_columns = [name for name, _ in columns]
    if columns:
        section_csv_file = os.path.join("sections", f"{section_tag}.csv")
        section_payloads[section_csv_file] = _compute_section_csv_payload(columns)

    return {
        "section_csv_file": section_csv_file,
        "section_csv_columns": section_csv_columns,
        "timestamp": section.timestamp,
        "baseline_in_queue": section.baseline_in_queue,
        "peaks_in_queue": section.peaks_in_queue,
        "peakinfo": section.peakinfo,
        "fit_result": fit_payload,
    }


def _dict_to_section(payload, param_dir, missing_files=None):
    section = Section()
    section_cols = {}
    section_csv_file = payload.get("section_csv_file")
    if section_csv_file is not None:
        full_csv = os.path.join(param_dir, section_csv_file)
        section_cols = _load_section_csv_columns(full_csv)
        if (section_cols == {}) and (not os.path.exists(full_csv)) and (missing_files is not None):
            missing_files.append(section_csv_file)

    if "x" in section_cols:
        section.x = section_cols.get("x")
    else:
        x_file = payload.get("x_file")
        if x_file is not None:
            section.x = _load_csv_array_legacy(os.path.join(param_dir, x_file))
        else:
            section.x = None if payload.get("x") is None else np.asarray(payload.get("x"), dtype=float)

    if "y_bgsub" in section_cols:
        section.y_bgsub = section_cols.get("y_bgsub")
    else:
        y_bgsub_file = payload.get("y_bgsub_file")
        if y_bgsub_file is not None:
            section.y_bgsub = _load_csv_array_legacy(os.path.join(param_dir, y_bgsub_file))
        else:
            section.y_bgsub = None if payload.get("y_bgsub") is None else np.asarray(payload.get("y_bgsub"), dtype=float)

    if "y_bg" in section_cols:
        section.y_bg = section_cols.get("y_bg")
    else:
        y_bg_file = payload.get("y_bg_file")
        if y_bg_file is not None:
            section.y_bg = _load_csv_array_legacy(os.path.join(param_dir, y_bg_file))
        else:
            section.y_bg = None if payload.get("y_bg") is None else np.asarray(payload.get("y_bg"), dtype=float)
    section.timestamp = payload.get("timestamp")
    section.baseline_in_queue = payload.get("baseline_in_queue", [])
    section.peaks_in_queue = payload.get("peaks_in_queue", [])
    section.peakinfo = payload.get("peakinfo", {})
    fit_payload = payload.get("fit_result")
    if isinstance(fit_payload, dict):
        best_col = fit_payload.get("best_fit_column", "fit_best")
        if best_col in section_cols:
            fit_payload["best_fit"] = np.asarray(section_cols.get(best_col), dtype=float).tolist()
        else:
            best_fit_file = fit_payload.get("best_fit_file")
            if best_fit_file is not None:
                old_arr = _load_csv_array_legacy(os.path.join(param_dir, best_fit_file))
                fit_payload["best_fit"] = np.asarray(old_arr, dtype=float).tolist()

        comps = {}
        comp_cols = fit_payload.get("component_columns", {})
        if isinstance(comp_cols, dict) and comp_cols:
            for comp_name, col_name in comp_cols.items():
                if col_name in section_cols:
                    comps[comp_name] = np.asarray(section_cols.get(col_name), dtype=float).tolist()
        elif section_cols:
            for col_name, values in section_cols.items():
                if col_name.startswith("comp_"):
                    comps[col_name[5:]] = np.asarray(values, dtype=float).tolist()
        else:
            comp_files = fit_payload.get("components_files", {})
            if isinstance(comp_files, dict) and comp_files:
                for k, rel in comp_files.items():
                    old_arr = _load_csv_array_legacy(os.path.join(param_dir, rel))
                    comps[k] = np.asarray(old_arr, dtype=float).tolist()
        if comps:
            fit_payload["components"] = comps
    section.fit_result = None if fit_payload is None else _FitResultLite(fit_payload)
    section.parameters = None
    section.fit_model = None
    return section


def _compute_np_payload(arr):
    buf = io.BytesIO()
    np.save(buf, arr)
    return buf.getvalue()


def _get_backup_index(path):
    if not os.path.exists(path):
        return {"format_family": FORMAT_FAMILY, "events": []}
    try:
        with open(path, "r", encoding="utf-8") as f:
            data = json.load(f)
            if not isinstance(data, dict):
                return {"format_family": FORMAT_FAMILY, "events": []}
            if "events" not in data or not isinstance(data.get("events"), list):
                data["events"] = []
            if "format_family" not in data:
                data["format_family"] = FORMAT_FAMILY
            return data
    except Exception:
        return {"format_family": FORMAT_FAMILY, "events": []}



def _make_backup_id(param_dir):
    index_path = os.path.join(param_dir, BACKUP_INDEX_FILE)
    index = _get_backup_index(index_path)
    used = set()
    for evt in index.get("events", []):
        bid = evt.get("id")
        if bid is None:
            continue
        try:
            used.add(int(str(bid)))
        except Exception:
            continue
    if os.path.isdir(param_dir):
        for entry in os.scandir(param_dir):
            if not entry.is_dir():
                continue
            try:
                used.add(int(str(entry.name)))
            except Exception:
                continue
    next_id = 0 if not used else (max(used) + 1)
    while next_id in used:
        next_id += 1
    return str(next_id)


def _collect_companion_files(param_dir):
    """
    Track legacy/session companion files in PARAM so their changes are included
    in backup snapshots as well.
    """
    suffixes = (
        ".poni", ".cake.npy", ".bg.chi", ".bgsub.chi", ".cakeformat",
        "smooth.chi",
    )
    rel_files = set()
    if not os.path.isdir(param_dir):
        return []
    for entry in os.scandir(param_dir):
        if (not entry.is_file()) or (not entry.name.endswith(suffixes)):
            continue
        rel_files.add(entry.name)
    return sorted(rel_files)


def _prepare_payloads(model, param_dir, ui_state=None):
    chi_root = model.chi_path
    cake_tth_file = None
    cake_azi_file = None
    cake_int_file = None
    mask_file = None
    if model.diff_img is not None:
        if model.diff_img.img_filename is not None:
            base = os.path.splitext(os.path.basename(model.diff_img.img_filename))[0]
        else:
            base = "cake"
        # Keep historical naming style used by write_temp_cakefiles.
        cake_tth_file = f"{base}.tth.cake.npy" if model.diff_img.tth_cake is not None else None
        cake_azi_file = f"{base}.azi.cake.npy" if model.diff_img.chi_cake is not None else None
        cake_int_file = f"{base}.int.cake.npy" if model.diff_img.intensity_cake is not None else None
        mask_file = f"{base}.mask.npy" if model.diff_img.mask is not None else None

    waterfall_payloads = {}
    waterfall_patterns = []
    for i, ptn in enumerate(model.waterfall_ptn):
        fallback_rel = None
        src = getattr(ptn, "_pkpo_original_fname", None) or getattr(ptn, "fname", None)
        if (src is not None) and os.path.exists(src):
            rel = os.path.join("waterfall", f"{i:04d}_{os.path.basename(src)}")
            try:
                waterfall_payloads[rel] = _file_bytes(src)
                fallback_rel = rel
            except Exception:
                fallback_rel = None
        waterfall_patterns.append(
            _serialize_pattern(
                ptn,
                chi_root,
                fallback_relpath=fallback_rel,
                force_abs_fname=True,
            )
        )

    session_data = {
        "schema": 1,
        "chi_path": ".",
        "jcpds_path": _relpath_or_abs(model.jcpds_path, chi_root),
        "poni": _relpath_or_abs(model.poni, chi_root),
        "base_pattern": _serialize_pattern(model.base_ptn, chi_root),
        "waterfall_patterns": waterfall_patterns,
        "diff_img": {
            "img_filename": _relpath_or_abs(getattr(model.diff_img, "img_filename", None), chi_root),
            "mask_file": mask_file,
            "cake_tth_file": cake_tth_file,
            "cake_azi_file": cake_azi_file,
            "cake_int_file": cake_int_file,
        },
        "current_section_index": None,
    }

    if model.current_section is not None:
        for i, sec in enumerate(model.section_lst):
            if sec.timestamp == model.current_section.timestamp:
                session_data["current_section_index"] = i
                break

    section_payloads = {}
    sections_data = {
        "schema": 1,
        "sections": [
            _section_to_dict(s, section_tag=f"sec_{i:04d}", section_payloads=section_payloads)
            for i, s in enumerate(model.section_lst)
        ],
        # Preserve in-progress/unsaved fit-section edits as part of state.
        "current_section": (
            _section_to_dict(
                model.current_section,
                section_tag="current_section",
                section_payloads=section_payloads,
            )
            if model.current_section is not None else None
        ),
    }
    jcpds_data = {
        "schema": 1,
        "saved_pressure": model.saved_pressure,
        "saved_temperature": model.saved_temperature,
        "phases": [_serialize_jcpds_item(j, chi_root) for j in model.jcpds_lst],
    }
    ui_data = {
        "schema": 1,
        "ui_state": ui_state or {},
    }
    return session_data, sections_data, jcpds_data, ui_data, section_payloads, waterfall_payloads


def _highlight_changed_files(changed_files):
    highlights = []
    if JCPDS_FILE in changed_files:
        highlights.append("JCPDS")
    if SECTIONS_FILE in changed_files:
        highlights.append("Fits")
    if SESSION_FILE in changed_files:
        highlights.append("Session")
    if UI_STATE_FILE in changed_files:
        highlights.append("UI")
    if not highlights and changed_files:
        highlights.append("Files")
    return highlights


def _safe_load_json(path):
    if not os.path.exists(path):
        return None
    try:
        with open(path, "r", encoding="utf-8") as f:
            return json.load(f)
    except Exception:
        return None


def _semantic_change_flags(snapshot_dir, session_data, sections_data, jcpds_data, ui_data):
    old_session = _safe_load_json(os.path.join(snapshot_dir, SESSION_FILE))
    old_sections = _safe_load_json(os.path.join(snapshot_dir, SECTIONS_FILE))
    old_jcpds = _safe_load_json(os.path.join(snapshot_dir, JCPDS_FILE))
    old_ui = _safe_load_json(os.path.join(snapshot_dir, UI_STATE_FILE))

    # Fits significance: compare only saved sections list, not current_section.
    old_sections_list = [] if old_sections is None else old_sections.get("sections", [])
    new_sections_list = sections_data.get("sections", [])
    # JCPDS significance with backward compatibility.
    old_phases = []
    if old_jcpds is not None:
        old_phases = old_jcpds.get("phases", old_jcpds.get("items", []))
    new_phases = jcpds_data.get("phases", [])

    return {
        "session": (old_session != session_data),
        "fits": (old_sections_list != new_sections_list),
        "jcpds": (old_phases != new_phases),
        "ui": (old_ui != ui_data),
    }


def _highlights_from_flags(flags, force_backup=False):
    highlights = []
    if flags.get("jcpds", False):
        highlights.append("JCPDS")
    if flags.get("fits", False):
        highlights.append("Fits")
    if flags.get("session", False):
        highlights.append("Session")
    if flags.get("ui", False):
        highlights.append("UI")
    if not highlights and force_backup:
        highlights.append("Snapshot")
    elif not highlights:
        highlights.append("none")
    return highlights


def _compute_state_hash(payload_map):
    h = hashlib.sha256()
    for rel_path in sorted(payload_map.keys()):
        h.update(rel_path.encode("utf-8"))
        h.update(b"\0")
        h.update(hashlib.sha256(payload_map[rel_path]).digest())
        h.update(b"\0")
    return h.hexdigest()


def _known_state_hashes(index):
    hashes = set()
    for ev in index.get("events", []):
        state_hash = ev.get("state_hash")
        if isinstance(state_hash, str) and state_hash != "":
            hashes.add(state_hash)
    return hashes


def _exclude_from_backup(rel_path):
    return False


def _cleanup_param_root(param_dir):
    keep_names = {BACKUP_INDEX_FILE}
    if not os.path.isdir(param_dir):
        return
    for entry in os.scandir(param_dir):
        if entry.name in keep_names:
            continue
        if entry.is_dir():
            continue
        try:
            os.remove(entry.path)
        except Exception:
            pass


def save_model_to_param(
        model, ui_state=None, reason="manual-save",
        force_backup=False, create_backup=True):
    if model is None or (not model.base_ptn_exist()):
        raise ValueError("Base pattern must exist before saving PARAM session.")
    param_dir = get_temp_dir(model.get_base_ptn_filename(), branch="-rampo")
    os.makedirs(param_dir, exist_ok=True)

    session_data, sections_data, jcpds_data, ui_data, section_payloads, waterfall_payloads = _prepare_payloads(
        model, param_dir=param_dir, ui_state=ui_state)
    index_path = os.path.join(param_dir, BACKUP_INDEX_FILE)
    index = _get_backup_index(index_path)
    latest_snapshot_dir = _resolve_snapshot_dir(param_dir)
    existing_created_at = None
    if latest_snapshot_dir is not None:
        existing_manifest_path = os.path.join(latest_snapshot_dir, MANIFEST_FILE)
        if os.path.exists(existing_manifest_path):
            try:
                existing_manifest = _load_json(existing_manifest_path)
                existing_created_at = existing_manifest.get("created_at")
            except Exception:
                existing_created_at = None

    manifest = {
        "format_family": FORMAT_FAMILY,
        "format_version": FORMAT_VERSION,
        # Keep created_at stable across saves so no-op saves do not look changed.
        "created_at": existing_created_at or datetime.datetime.now().isoformat(timespec="seconds"),
        "files": {
            "session": SESSION_FILE,
            "sections": SECTIONS_FILE,
            "jcpds": JCPDS_FILE,
            "ui_state": UI_STATE_FILE,
            "backup_index": os.path.join("..", BACKUP_INDEX_FILE),
        },
    }

    payload_map = {
        SESSION_FILE: json.dumps(session_data, indent=2, sort_keys=True, default=_json_default).encode("utf-8"),
        SECTIONS_FILE: json.dumps(sections_data, indent=2, sort_keys=True, default=_json_default).encode("utf-8"),
        # Keep insertion order for readability:
        # phase fields first, DiffLines after; and h/k/l grouped in each line.
        JCPDS_FILE: json.dumps(jcpds_data, indent=2, sort_keys=False, default=_json_default).encode("utf-8"),
        UI_STATE_FILE: json.dumps(ui_data, indent=2, sort_keys=True, default=_json_default).encode("utf-8"),
        MANIFEST_FILE: json.dumps(manifest, indent=2, sort_keys=True, default=_json_default).encode("utf-8"),
    }
    payload_map.update(section_payloads)
    payload_map.update(waterfall_payloads)

    # Build companion files in the top-level param dir first, then move
    # the resulting payloads into the numbered snapshot only.
    if model.base_ptn is not None:
        try:
            model.base_ptn.write_temporary_bgfiles(temp_dir=param_dir)
        except Exception:
            pass
        try:
            model.base_ptn.write_temporary_processed_file(temp_dir=param_dir)
        except Exception:
            pass
    if model.waterfall_ptn:
        for ptn in model.waterfall_ptn:
            try:
                ptn.write_temporary_bgfiles(temp_dir=param_dir)
            except Exception:
                pass

    # Keep cake naming/style consistent with existing temp npy files.
    if model.diff_img is not None:
        if model.diff_img.tth_cake is not None and model.diff_img.chi_cake is not None and model.diff_img.intensity_cake is not None:
            try:
                model.diff_img.write_temp_cakefiles(temp_dir=param_dir)
            except Exception:
                pass
        diff_info = session_data.get("diff_img", {})
        for key in ("cake_tth_file", "cake_azi_file", "cake_int_file"):
            rel = diff_info.get(key)
            if rel is None:
                continue
            full = os.path.join(param_dir, rel)
            if os.path.exists(full):
                payload_map[rel] = _file_bytes(full)
        if model.diff_img.mask is not None:
            rel = diff_info.get("mask_file")
            if rel is not None:
                payload_map[rel] = _compute_np_payload(model.diff_img.mask)

    # Add processed and BG chi files to tracked payload so they are backed up/restored.
    for ptn in [model.base_ptn] + list(model.waterfall_ptn):
        if ptn is None:
            continue
        processed_name = _processed_temp_name(ptn)
        bgsub_name, bg_name = _bg_temp_names(ptn)
        for rel in (processed_name, bgsub_name, bg_name):
            full = os.path.join(param_dir, rel)
            if os.path.exists(full):
                payload_map[rel] = _file_bytes(full)
    smooth_name = _smoothed_temp_name()
    smooth_full = os.path.join(param_dir, smooth_name)
    if os.path.exists(smooth_full):
        payload_map[smooth_name] = _file_bytes(smooth_full)

    # Include companion files (e.g., poni/cakeformat/cake npy/bg chi) in
    # change detection + backups, preserving historical PARAM usage.
    for rel in _collect_companion_files(param_dir):
        full = os.path.join(param_dir, rel)
        if os.path.exists(full):
            payload_map[rel] = _file_bytes(full)

    changed_files_all = []
    compare_root = latest_snapshot_dir if latest_snapshot_dir is not None else param_dir
    for rel_path, new_bytes in payload_map.items():
        full_path = os.path.join(compare_root, rel_path)
        if _file_differs_from_payload(full_path, new_bytes):
            changed_files_all.append(rel_path)

    backup_payload_map = {
        rel_path: payload
        for rel_path, payload in payload_map.items()
        if not _exclude_from_backup(rel_path)
    }
    changed_files = [p for p in changed_files_all if p in backup_payload_map]
    delta_files = list(changed_files)
    semantic_flags = _semantic_change_flags(
        compare_root, session_data, sections_data, jcpds_data, ui_data)
    state_hash = _compute_state_hash(backup_payload_map)

    backup_id = None
    backup_root = None
    known_hashes = _known_state_hashes(index)
    state_already_known = state_hash in known_hashes

    create_backup_event = create_backup and ((changed_files != []) or force_backup)
    if create_backup_event and state_already_known and (not force_backup):
        # Avoid duplicate backups for states that already exist in history.
        create_backup_event = False

    snapshot_files = sorted(backup_payload_map.keys())
    if create_backup_event:
        backup_id = _make_backup_id(param_dir)
        backup_root = os.path.join(param_dir, backup_id)
        for rel_path in snapshot_files:
            dst = os.path.join(backup_root, rel_path)
            _atomic_write_bytes(dst, backup_payload_map[rel_path])

    if create_backup_event:
        highlights = _highlights_from_flags(semantic_flags, force_backup=force_backup)
        index["events"].append(
            {
                "id": backup_id,
                "timestamp": datetime.datetime.now().isoformat(timespec="seconds"),
                "reason": reason,
                "changed_files": delta_files,
                "snapshot_files": snapshot_files,
                "snapshot_mode": "full",
                "highlights": highlights,
                "state_hash": state_hash,
            }
        )
        index["latest_backup_id"] = backup_id
    _atomic_write_json(index_path, index)

    _cleanup_param_root(param_dir)

    return SaveResult(
        param_dir=param_dir,
        manifest_path=os.path.join(backup_root, MANIFEST_FILE) if backup_root is not None else "",
        backup_id=backup_id,
        changed_files=changed_files_all,
    )


def _load_json(path):
    with open(path, "r", encoding="utf-8") as f:
        return json.load(f)


def is_new_param_folder(param_dir):
    try:
        index = _get_backup_index(os.path.join(param_dir, BACKUP_INDEX_FILE))
    except Exception:
        return False
    if index.get("format_family") != FORMAT_FAMILY:
        return False
    events = index.get("events", [])
    if not isinstance(events, list) or len(events) == 0:
        return False
    latest_id = index.get("latest_backup_id")
    if latest_id in (None, ""):
        latest_id = events[-1].get("id")
    if latest_id in (None, ""):
        return False
    return os.path.isdir(os.path.join(param_dir, str(latest_id)))


def list_backup_events(param_dir):
    index_path = os.path.join(param_dir, BACKUP_INDEX_FILE)
    index = _get_backup_index(index_path)
    return index.get("events", [])


def _resolve_snapshot_dir(param_dir, event_id=None, event_index=None):
    events = list_backup_events(param_dir)
    if not events:
        return None
    if event_index is not None:
        if (event_index < 0) or (event_index >= len(events)):
            return None
        target_pos = int(event_index)
    else:
        if event_id is None:
            index = _get_backup_index(os.path.join(param_dir, BACKUP_INDEX_FILE))
            event_id = index.get("latest_backup_id")
            if event_id in (None, "") and events:
                event_id = events[-1].get("id")
        ids = [e.get("id") for e in events]
        if event_id not in ids:
            return None
        target_pos = max(i for i, _id in enumerate(ids) if _id == event_id)
    target_evt = events[target_pos]
    snap_dir = os.path.join(param_dir, str(target_evt.get("id", "")))
    if not os.path.isdir(snap_dir):
        return None
    return snap_dir


def restore_to_backup_event(param_dir, event_id=None, event_index=None):
    return _resolve_snapshot_dir(param_dir, event_id=event_id, event_index=event_index) is not None


def load_model_from_param(model, base_chi_file, backup_event_id=None, backup_event_index=None):
    chi_root = os.path.dirname(base_chi_file)
    param_dir = get_temp_dir(base_chi_file, branch="-rampo")
    if not is_new_param_folder(param_dir):
        return False, {"reason": "missing-or-invalid-manifest", "param_dir": param_dir}

    snapshot_dir = _resolve_snapshot_dir(
        param_dir,
        event_id=backup_event_id,
        event_index=backup_event_index,
    )
    if snapshot_dir is None:
        return False, {
            "reason": "invalid-backup-id",
            "param_dir": param_dir,
            "backup_id": backup_event_id,
            "backup_index": backup_event_index,
        }

    manifest = _load_json(os.path.join(snapshot_dir, MANIFEST_FILE))
    files = manifest.get("files", {})
    session_data = _load_json(os.path.join(snapshot_dir, files.get("session", SESSION_FILE)))
    sections_data = _load_json(os.path.join(snapshot_dir, files.get("sections", SECTIONS_FILE)))
    jcpds_data = _load_json(os.path.join(snapshot_dir, files.get("jcpds", JCPDS_FILE)))
    ui_data = _load_json(os.path.join(snapshot_dir, files.get("ui_state", UI_STATE_FILE)))

    model.base_ptn = _load_pattern(session_data.get("base_pattern"), chi_root, snapshot_dir)
    model.waterfall_ptn = [
        _load_pattern(p, chi_root, snapshot_dir)
        for p in session_data.get("waterfall_patterns", [])
    ]
    fallback_waterfall = []
    for ptn in model.waterfall_ptn:
        if bool(getattr(ptn, "_pkpo_fallback_in_use", False)):
            fallback_waterfall.append(
                os.path.basename(
                    getattr(ptn, "_pkpo_original_fname", None) or getattr(ptn, "fname", "") or ""
                )
            )
    phase_payloads = jcpds_data.get("phases", jcpds_data.get("items", []))
    model.jcpds_lst = [_load_jcpds_item(p, chi_root) for p in phase_payloads]
    missing_section_csv_files = []
    loaded_sections = []
    old_to_new_idx = {}
    for old_idx, sec_payload in enumerate(sections_data.get("sections", [])):
        sec_obj = _dict_to_section(sec_payload, snapshot_dir, missing_files=missing_section_csv_files)
        x_arr = getattr(sec_obj, "x", None)
        if (x_arr is None) or (len(x_arr) == 0):
            continue
        old_to_new_idx[old_idx] = len(loaded_sections)
        loaded_sections.append(sec_obj)
    model.section_lst = loaded_sections

    current_idx = session_data.get("current_section_index")
    current_payload = sections_data.get("current_section")
    model.current_section = None
    if current_payload is not None:
        current_section = _dict_to_section(
            current_payload, snapshot_dir, missing_files=missing_section_csv_files)
        if (getattr(current_section, "x", None) is None) or (len(current_section.x) == 0):
            current_section = None
        mapped_idx = old_to_new_idx.get(current_idx) if isinstance(current_idx, int) else None
        if (current_section is not None) and isinstance(mapped_idx, int) and (0 <= mapped_idx < len(model.section_lst)):
            listed = model.section_lst[mapped_idx]
            if getattr(listed, "timestamp", None) == getattr(current_section, "timestamp", None):
                model.current_section = listed
            else:
                model.current_section = current_section
        else:
            model.current_section = current_section
    elif isinstance(current_idx, int):
        mapped_idx = old_to_new_idx.get(current_idx)
        if isinstance(mapped_idx, int) and (0 <= mapped_idx < len(model.section_lst)):
            model.current_section = model.section_lst[mapped_idx]

    model.saved_pressure = jcpds_data.get(
        "saved_pressure", session_data.get("saved_pressure", model.saved_pressure))
    model.saved_temperature = jcpds_data.get(
        "saved_temperature", session_data.get("saved_temperature", model.saved_temperature))
    model.chi_path = chi_root
    model.jcpds_path = _resolve_path(session_data.get("jcpds_path"), chi_root) or ""
    model.poni = _resolve_path(session_data.get("poni"), chi_root)

    diff_img_info = session_data.get("diff_img", {})
    img_filename = _resolve_path(diff_img_info.get("img_filename"), chi_root)
    has_any = any(
        diff_img_info.get(k) is not None
        for k in ("mask_file", "cake_tth_file", "cake_azi_file", "cake_int_file")
    ) or (img_filename is not None)
    if has_any:
        diff = DiffImg()
        diff.img_filename = img_filename
        mask_file = diff_img_info.get("mask_file")
        cake_tth_file = diff_img_info.get("cake_tth_file")
        cake_azi_file = diff_img_info.get("cake_azi_file")
        cake_int_file = diff_img_info.get("cake_int_file")
        if mask_file is not None:
            f = os.path.join(snapshot_dir, mask_file)
            if os.path.exists(f):
                diff.mask = np.load(f, allow_pickle=False)
        if cake_tth_file is not None:
            f = os.path.join(snapshot_dir, cake_tth_file)
            if os.path.exists(f):
                diff.tth_cake = np.load(f, allow_pickle=False)
        if cake_azi_file is not None:
            f = os.path.join(snapshot_dir, cake_azi_file)
            if os.path.exists(f):
                diff.chi_cake = np.load(f, allow_pickle=False)
        if cake_int_file is not None:
            f = os.path.join(snapshot_dir, cake_int_file)
            if os.path.exists(f):
                diff.intensity_cake = np.load(f, allow_pickle=False)
        model.diff_img = diff
        try:
            if hasattr(model.diff_img, "apply_excitation_wavelength") and \
                    hasattr(model.base_ptn, "wavelength"):
                model.diff_img.apply_excitation_wavelength(model.base_ptn.wavelength)
        except Exception:
            pass
    else:
        model.diff_img = None

    return True, {
        "param_dir": param_dir,
        "manifest": os.path.join(snapshot_dir, MANIFEST_FILE),
        "ui_state": ui_data.get("ui_state", {}),
        "missing_section_csv_files": sorted(set(missing_section_csv_files)),
        "fallback_waterfall_files": sorted(set([x for x in fallback_waterfall if x])),
        "category_presence": {
            "backup_information": (len(list_backup_events(param_dir)) > 0),
            "jcpds": (len(phase_payloads) > 0),
            "pressure": (("saved_pressure" in jcpds_data) or ("saved_pressure" in session_data)),
            "temperature": (("saved_temperature" in jcpds_data) or ("saved_temperature" in session_data)),
            "spectrum_smoothing": (ui_data.get("ui_state", {}).get("spectrum", {}) != {}),
            "cake_z_scale": (ui_data.get("ui_state", {}).get("cake", {}) != {}),
            "ccd_roi": (ui_data.get("ui_state", {}).get("ccd_roi", {}) != {}),
            "background": (ui_data.get("ui_state", {}).get("background", {}) != {}),
            "waterfall_list": (len(session_data.get("waterfall_patterns", [])) > 0),
            "poni": (session_data.get("poni") not in (None, "")),
            "fits_information": (
                (len(sections_data.get("sections", [])) > 0) or
                (sections_data.get("current_section") is not None)
            ),
        },
    }


def load_section_from_param(base_chi_file, section_index):
    """
    Load one saved section by index from PARAM session files.
    Returns Section object on success, or None if unavailable/invalid.
    """
    if base_chi_file is None:
        return None
    param_dir = get_temp_dir(base_chi_file, branch="-rampo")
    if not is_new_param_folder(param_dir):
        return None
    try:
        snapshot_dir = _resolve_snapshot_dir(param_dir)
        if snapshot_dir is None:
            return None
        manifest = _load_json(os.path.join(snapshot_dir, MANIFEST_FILE))
        files = manifest.get("files", {})
        sections_data = _load_json(os.path.join(snapshot_dir, files.get("sections", SECTIONS_FILE)))
        sections = sections_data.get("sections", [])
        idx = int(section_index)
        if (idx < 0) or (idx >= len(sections)):
            return None
        return _dict_to_section(sections[idx], snapshot_dir)
    except Exception:
        return None
